""" Interface for transformers' ValueMakers.
"""
import logging
import abc

from . import errormanager
from . import exceptions

logger = logging.getLogger("ontoweaver")
class ValueMaker(errormanager.ErrorManager, metaclass=abc.ABCMeta):
    """
    Interface for value makers that are used to define the extraction logic for each of the transformer classes.
    """
    def __init__(self, raise_errors: bool = True):
        """
        Initialize the ValueMaker with the option to raise errors.

        Args:
            raise_errors (bool): if True, will raise an exception when an error is encountered, else, will log the error and try to proceed.
        """

        super().__init__(raise_errors)

    @abc.abstractmethod
    def __call__(self, columns, row, i):
        """
        Select the value from the columns and row based on the logic defined in the subclass.
        Args:
            columns: The columns of the transformer instance.
            row: The current row of the dataframe.
            i: The current index of the row.

        Returns:
            The selected value.
        """
        raise NotImplementedError("The call method must be implemented in a subclass.")



import logging

import yaml

import pandas as pd

import biocypher
import ontoweaver

logger =logging.getLogger("ontoweaver")

def test_node_type_fusion():

    data_file = "tests/test_node_type_fusion/test.csv"
    
    # Test of the most generic common subtype fusion 

    filename_to_mapping = {data_file : "tests/test_node_type_fusion/mapping.yaml"}
    logger.debug("Load data...")
        
    nodes, edges = ontoweaver.extract(filename_to_mapping)

    bc_nodes, bc_edges = ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges)

    on_ID = ontoweaver.serialize.ID()
    congregater = ontoweaver.congregate.Nodes(on_ID)
    for n in congregater(bc_nodes):
        pass

    logger.debug('Initialize Biocypher instance...')

    bc = biocypher.BioCypher(
        biocypher_config_path = "tests/test_node_type_fusion/biocypher_config.yaml",
        schema_config_path = "tests/test_node_type_fusion/schema_config.yaml"
    )

    as_keys  = ontoweaver.merge.string.UseKey()
    as_sub_type = ontoweaver.merge.string.CommonSubType(bc._get_ontology())
    in_lists = ontoweaver.merge.dictry.Append()
    fuser = ontoweaver.fuse.Members(ontoweaver.base.Node,
            merge_ID    = as_keys,
            merge_label = as_sub_type,
            merge_prop  = in_lists,
        )

    fusioner = ontoweaver.fusion.Reduce(fuser)
    fusioned = fusioner(congregater)

    logger.debug("Fusioned items:")
    fused = []
    f_str = []
    for f in fusioned:
        logger.debug("  "+repr(f))
        fused.append(f)
        f_str.append(f._label)

    nb_persons = 0
    with open(data_file) as f:
        nb_persons = sum(1 for line in f) - 1
    assert nb_persons == len(fused)
    
    data_df = pd.read_csv(data_file)
    nb_male = len(data_df[data_df['genre'] == 'male'])
    nb_female = len(data_df[data_df['genre'] == 'female'])

    nb_fused_male = f_str.count("male")
    nb_fused_female = f_str.count("female")

    assert nb_male == nb_fused_male
    assert nb_female == nb_fused_female

    # Test of the most generic common subtype fusion with multiple inheritance (intersex example)

    filename_to_mapping = {data_file : "tests/test_node_type_fusion/mapping_only_mothers.yaml"}
    logger.debug("Load data...")
        
    nodes, edges = ontoweaver.extract(filename_to_mapping)

    bc_nodes, bc_edges = ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges)

    on_ID = ontoweaver.serialize.ID()
    congregater = ontoweaver.congregate.Nodes(on_ID)
    for n in congregater(bc_nodes):
        pass

    logger.debug('Initialize Biocypher instance...')

    bc = biocypher.BioCypher(
        biocypher_config_path = "tests/test_node_type_fusion/biocypher_config.yaml",
        schema_config_path = "tests/test_node_type_fusion/schema_config.yaml"
    )

    as_keys  = ontoweaver.merge.string.UseKey()
    as_sub_type = ontoweaver.merge.string.CommonSubType(bc._get_ontology())
    in_lists = ontoweaver.merge.dictry.Append()
    fuser = ontoweaver.fuse.Members(ontoweaver.base.Node,
            merge_ID    = as_keys,
            merge_label = as_sub_type,
            merge_prop  = in_lists,
        )

    fusioner = ontoweaver.fusion.Reduce(fuser)
    fusioned = fusioner(congregater)

    logger.debug("Fusioned items:")
    fused = []
    f_str = []
    for f in fusioned:
        logger.debug("  "+repr(f))
        fused.append(f)
        f_str.append(f._label)

    nb_persons = 0
    with open(data_file) as f:
        nb_persons = sum(1 for line in f) - 1
    assert nb_persons == len(fused)
    
    data_df = pd.read_csv(data_file)
    nb_female = len(data_df[data_df['genre'] == 'female'])
    nb_intersex = int(data_df.notna().sum()['children'])
    nb_male = len(data_df[data_df['genre'] == 'male']) - nb_intersex

    nb_fused_male = f_str.count("male")
    nb_fused_female = f_str.count("female")
    nb_fused_intersex = f_str.count("intersex")

    assert nb_male == nb_fused_male
    assert nb_female == nb_fused_female
    assert nb_intersex == nb_fused_intersex



    # Test of the most specific common supertype fusion 

    filename_to_mapping = {data_file : "tests/test_node_type_fusion/mapping_only_mothers.yaml"}
    logger.debug("Load data...")
        
    nodes, edges = ontoweaver.extract(filename_to_mapping)

    bc_nodes, bc_edges = ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges)

    on_ID = ontoweaver.serialize.ID()
    congregater = ontoweaver.congregate.Nodes(on_ID)
    for n in congregater(bc_nodes):
        pass

    logger.debug('Initialize Biocypher instance...')

    bc = biocypher.BioCypher(
        biocypher_config_path = "tests/test_node_type_fusion/biocypher_config.yaml",
        schema_config_path = "tests/test_node_type_fusion/schema_config.yaml"
    )
    # bc.show_ontology_structure()

    as_keys  = ontoweaver.merge.string.UseKey()
    as_sub_type = ontoweaver.merge.string.CommonSuperType(bc._get_ontology())
    in_lists = ontoweaver.merge.dictry.Append()
    fuser = ontoweaver.fuse.Members(ontoweaver.base.Node,
            merge_ID    = as_keys,
            merge_label = as_sub_type,
            merge_prop  = in_lists,
        )

    fusioner = ontoweaver.fusion.Reduce(fuser)
    fusioned = fusioner(congregater)

    logger.debug("Fusioned items:")
    fused = []
    f_str = []
    for f in fusioned:
        logger.debug("  "+repr(f))
        fused.append(f)
        f_str.append(f._label)

    nb_persons = 0
    with open(data_file) as f:
        nb_persons = sum(1 for line in f) - 1
    assert nb_persons == len(fused)
    
    data_df = pd.read_csv(data_file)
    nb_female = len(data_df[data_df['genre'] == 'female'])
    nb_person = data_df.notna().sum()['children']
    nb_male = len(data_df[data_df['genre'] == 'male']) - nb_person

    nb_fused_male = f_str.count("male")
    nb_fused_female = f_str.count("female")
    nb_fused_person = f_str.count("person")

    assert nb_male == nb_fused_male
    assert nb_female == nb_fused_female
    assert nb_person == nb_fused_person

if __name__ == "__main__":
    test_node_type_fusion()

    



def test_final_type():
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "final_type"

    expected_nodes = [
        ('chair:aaaaaa', 'aaaaaa', {'localisation': 'Peterkitchen', 'blabla': 'blabla', 'source_columns': 'furniture'}),
        ('kitchen:eeeeee', 'eeeeee', {'blabla': 'blabla', 'source_columns': 'localisation'}),
        ('Paul:cccccc', 'cccccc', {'blabla': 'blabla', 'source_columns': 'name'}),
        ('bathroom:eeeeee', 'eeeeee', {'blabla': 'blabla', 'source_columns': 'localisation'}),
        ('Mary:dddddd', 'dddddd', {'blabla': 'blabla', 'source_columns': 'name'}),
        ('sofa:aaaaaa', 'aaaaaa', {'localisation': 'Paulbathroom', 'blabla': 'blabla', 'source_columns': 'furniture'}),
        ('fridge:aaaaaa', 'aaaaaa', {'localisation': 'Marykitchen', 'blabla': 'blabla', 'source_columns': 'furniture'}),
        ('Peter:dddddd', 'dddddd', {'blabla': 'blabla', 'source_columns': 'name'})
    ]

    expected_edges = [
        ('(chair:aaaaaa)--[has_localisation]->(kitchen:eeeeee)', 'chair:aaaaaa', 'kitchen:eeeeee', 'has_localisation', {'blabla': 'blabla'}),
        ('(chair:aaaaaa)--[will_not_sit]->(Peter:dddddd)', 'chair:aaaaaa', 'Peter:dddddd', 'will_not_sit', {'blabla': 'blabla'}),
        ('(fridge:aaaaaa)--[will_not_sit]->(Mary:dddddd)', 'fridge:aaaaaa', 'Mary:dddddd', 'will_not_sit', {'blabla': 'blabla'}),
        ('(sofa:aaaaaa)--[will_sit]->(Paul:cccccc)', 'sofa:aaaaaa', 'Paul:cccccc', 'will_sit', {'blabla': 'blabla'}),
        ('(sofa:aaaaaa)--[has_localisation]->(bathroom:eeeeee)', 'sofa:aaaaaa', 'bathroom:eeeeee', 'has_localisation', {'blabla': 'blabla'}),
        ('(fridge:aaaaaa)--[has_localisation]->(kitchen:eeeeee)', 'fridge:aaaaaa', 'kitchen:eeeeee', 'has_localisation', {'blabla': 'blabla'})
    ]

    data_mapping = {f"tests/{directory_name}/data.csv" : f"tests/{directory_name}/mapping.yaml" }

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix")

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges), reconciliate_sep=",")

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_final_type()



import logging
import yaml
import io
import pandas as pd

import ontoweaver

def test_transformer_boolean():

    logging.debug("Load data...")

    # Do not add newlines or spaces here
    # or else the parsing will be wrong.
    data = """Patient,Variant,Source
P1,V1-1,S0
P1,V1-2,S1
P2,V2-1,S2
P2,V2-2,S3"""
    csv = io.StringIO(data)
    table = pd.read_csv(csv)

    logging.debug("Load mappings...")

    mapping = """
row:
    map:
        column: Variant
        to_subject: variant
transformers:
    - boolean:
        column: Source
        to_object: source_quality
        via_relation: has_source_quality
        consider_true:
            - S0
            - S1
        consider_false:
            - S2
            - S3
        output_true: good
        output_false: not_good
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="none")

    for node in nodes:
        print(node.as_tuple())
        if node.as_tuple()[1] == "source_quality":
            assert node.as_tuple()[0] in ["good", "not_good"]


if __name__ == "__main__":
    test_transformer_boolean()




import logging
import subprocess
import os

def test_validate_input():
    logging.basicConfig(level=logging.DEBUG)

    logging.debug(f"From: {os.getcwd()}")
    cmd = "ontoweave ./tests/validate_input/data.csv:./tests/validate_input/mapping.yaml --validate-only"

    logging.debug(f"Run: {cmd}")

    result = subprocess.run(cmd.split(), capture_output=True, text=True, check=False)

    # Check if the return code is 76 - indicating the validation has detected an error. This is the expected behavior,
    # so test passes in this case.
    if result.returncode == 76:
        logging.info("Test passed: Command returned the expected exit code 76.")
    else:
        raise Exception(f"Test failed: Expected exit code 76 but got {result.returncode}.")

if __name__ == "__main__":
    test_validate_input()



""" Class for input and output data validation.
"""
import math
import logging
import warnings

import pandas as pd
import pandera.pandas as pa
from pandera import Column, Check

from . import errormanager
from . import exceptions

logger = logging.getLogger("ontoweaver")


class Validator(errormanager.ErrorManager):
    """Class used for data validation against a schema. The class uses the Pandera package to validate the data frame."""

    def __init__(self, validation_rules: pa.DataFrameSchema = None, raise_errors = True):
        """Constructor for the Validator class.

        Args:
            validation_rules: The schema used for validation.
        """
        super().__init__(raise_errors)

        self.validation_rules: pa.DataFrameSchema = validation_rules

        if self.validation_rules:
            self.make_raise_warnings(self.validation_rules.columns)

        # Repeated validation error messages.
        self.messages = {}

        self.dtypes = {
            'int64': int,
            'float64': float
        }


    def error(self, msg, section = None):
        super().error(msg, section = section, exception = exceptions.DataValidationError)
        err = self.messages.get(msg, {"count": 0, "section": section})
        err["count"] += 1
        self.messages[msg] = err


    def __call__(self, df, section = None):
        """
        Validate the data frame against the schema.

        Args:
            df: The data frame to validate.

        Returns:
            bool: True if the data frame is valid, Otherwise raise an pa.errors.SchemaError.
        """
        if self.validation_rules:
            logger.debug(f"Input validation rules: on {len(self.validation_rules.columns)} columns:")
            for col in self.validation_rules.columns:
                dtype = self.validation_rules.columns[col].dtype
                logger.debug(f"\t{col}: {dtype}")
                if str(dtype) in self.dtypes:
                    logger.debug(f"\t\tConvert to {self.dtypes[str(dtype)]}")
                    df[col] = df[col].astype(self.dtypes[str(dtype)])

            # May raise a pa.errors.SchemaError which will be catched in caller.
            # Generally base.Transformer.label_maker(...), so that
            # we know which transformer deals the issue.
            # try:
            #     self.validation_rules.validate(df)
            #     return True
            # except pa.errors.SchemaError as e:
            #     self.error(str(e), section = section)
            #     return False
            with warnings.catch_warnings(record = True) as caught:
                warnings.simplefilter("always")
                self.validation_rules.validate(df)
                for warn in caught:
                    self.error(str(warn.message), section = section)

                if caught:
                    return False
                else:
                    return True
        else:
            logger.warning("No schema provided for data validation.")
            return True


    def make_raise_warnings(self, rules = None):
        if not rules:
            rules = self.validation_rules.columns

        # Add the raise_warning predicate for all rules.
        for k in rules:
            for i,c in enumerate(rules[k].checks):
                rules[k].checks[i].raise_warning = True


class InputValidator(Validator):
    """Class used for input data validation against a schema. The class uses the Pandera package to validate the data frame."""

    def __init__(self, validation_rules: pa.DataFrameSchema = None, raise_errors = True):
        """Constructor for the InputValidator class.

        Args:
            validation_rules: The schema used for validation.
        """

        super().__init__(validation_rules, raise_errors)

    def __call__(self, df):
        """
        Validate the data frame against the schema.

        Args:
            df: The data frame to validate.

        Returns:
            bool: True if the data frame is valid, False otherwise.
        """

        return super().__call__(df, section = self.__class__.__name__)


no_validation_rules: pa.DataFrameSchema = pa.DataFrameSchema({})

default_validation_rules: pa.DataFrameSchema = pa.DataFrameSchema({
            "cell_value": Column(
                pa.Object,
                checks=[
                    Check(
                        # We pass a lambda function which checks if the value is numeric or non-numeric NaN,
                        # or an empty string. This is meant to be the default behavior of an instance of the
                        # OutputValidator class.
                        lambda s: (
                              s.apply(lambda x: pd.api.types.is_numeric_dtype(type(x))
                                                and not pd.isna(x))
                            | s.apply(lambda x: not pd.api.types.is_numeric_dtype(type(x))
                                                and str(x).lower() != "nan"
                                                and x != "")
                        ),
                        error="Invalid value detected",
                        raise_warning = True,
                    )
                ],

                # We additionally set that a cell value cannot be null.
                nullable=False
            )
    })


class OutputValidator(Validator):
    """Class used for transformer output data validation against a schema.
    The schema contains some general rules for the output data frame expected by default. The user can add additional rules
    to the schema for each transformer type used on each column. The class uses the Pandera package to validate the data frame."""

    def __init__(self,
                validation_rules = default_validation_rules,
            raise_errors = True):
        """Constructor for the OutputValidator class. By default, the output validator is instantiated with a schema that
        checks for numeric and non - numeric NaN values, and empty strings.

        Args:
            validation_rules: The schema used for validation.
        """
        super().__init__(validation_rules, raise_errors)


    def __call__(self, df):
        """
        Validate the data frame against the schema.

        Args:
            df: The data frame to validate.

        Returns:
            bool: True if the data frame is valid, False otherwise.
        """
        return super().__call__(df, section = self.__class__.__name__)


    def update_rules(self, new_rules):
        """Update the validation schema with additional rules.

        Args:
            new_rules: The additional rules to be added to the schema.
        """
        if not isinstance(new_rules, pa.DataFrameSchema):
            self.error("new_rules must be a Pandera DataFrameSchema instance.", section = self.__class__.__name__)


        # Merge the existing rules with the new ones
        merged_rules = self.validation_rules.columns.copy() if self.validation_rules else {}

        merged_rules.update(new_rules.columns)

        self.make_raise_warnings(merged_rules)

        # Update the validation rules
        self.validation_rules = pa.DataFrameSchema(merged_rules)


class SimpleOutputValidator(Validator):

    def __init__(self, validation_rules = None, raise_errors = True):
        """Constructor for the SimpleValidator class. This class is used in case of absence of any validation rules outside
        of checking for none-types and empty strings. The class is used to bypass the use of the Pandera package for validation,
        which saves a lot of computational time when big datasets are used.

        Args:
            validation_rules: The schema used for validation.
        """
        super().__init__(validation_rules, raise_errors)


    def __call__(self, val):
        if pd.api.types.is_numeric_dtype(type(val)):
            if math.isnan(val) or val == float("nan"):
                return False

        elif str(val).lower() == "nan" or val == "":
            return False

        return True


class SkipValidator(Validator):
    """Class used for skipping validation. This class is used to skip validation for specific transformers.
    We implement the SkipValidator class in order to keep the validation logic in the same place as the other validators,
    and keep the code in base.py uniform for all validation options."""

    def __init__(self, validation_rules = None, raise_errors = True):
        """Constructor for the SkipValidator class.

        Args:
            validation_rules: The schema used for validation.
        """
        super().__init__(validation_rules, raise_errors)
        logger.debug("Instantiate a `SkipValidator`, some output validation will be skipped." \
                     " This could result in some empty or `nan` nodes in your knowledge graph." \
                     " To enable output validation set validate_output to True.")


    def __call__(self, val):
        return True




def test_match_type_from_column():
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "match_type_from_column"

    expected_nodes = [
        ('fridge:kitchen_furniture', 'kitchen_furniture', {'localisation': 'Marykitchen', 'blabla': 'blabla', 'source_columns': 'furniture'}),
        ('Peter:person', 'person', {'blabla': 'blabla', 'source_columns': 'name'}),
        ('sofa:rest_of_house_furniture', 'rest_of_house_furniture', {'localisation': 'Paulbathroom', 'blabla': 'blabla', 'source_columns': 'furniture'}),
        ('chair:kitchen_furniture', 'kitchen_furniture', {'localisation': 'Peterkitchen', 'blabla': 'blabla', 'source_columns': 'furniture'}),
        ('Paul:person', 'person', {'blabla': 'blabla', 'source_columns': 'name'}),
        ('Mary:person', 'person', {'blabla': 'blabla', 'source_columns': 'name'}),
    ]

    expected_edges = [
        ('', 'sofa:rest_of_house_furniture', 'Paul:person', 'will_sit', {'blabla': 'blabla'}),
        ('', 'chair:kitchen_furniture', 'Peter:person', 'will_not_sit', {'blabla': 'blabla'}),
        ('', 'fridge:kitchen_furniture', 'Mary:person', 'will_not_sit', {'blabla': 'blabla'}),
    ]

    data_mapping = {f"tests/{directory_name}/data.csv": f"tests/{directory_name}/mapping.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix")

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges), reconciliate_sep=",")

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_match_type_from_column()



import logging
import yaml
import pandas as pd
import ontoweaver
from ontoweaver import exceptions

class user_transformer(ontoweaver.base.Transformer):

    class ValueMaker(ontoweaver.make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:
                if key not in row:
                    self.error(f"Column '{key}' not found in data", section="map.call",
                               exception=exceptions.TransformerDataError)
                yield row[key]

    def __init__(self, properties_of, value_maker = ValueMaker(), label_maker = None, branching_properties=None, columns=None, **kwargs):
        super().__init__(properties_of, value_maker, label_maker, branching_properties, columns, **kwargs)

    def __call__(self, row, i):

        for value in self.value_maker(self.columns, row, i):
            yield self.create(value, row)


def test_transformer_user():
    # Add the passed transformer to the list available to OntoWeaver.
    ontoweaver.transformer.register(user_transformer)

    directory_name = "simplest"

    logging.debug("Load mapping...")
    yaml_mapping = """
    row:
        rowIndex:
            to_subject: variant
    transformers:
        - user_transformer:
            columns:
                - patient
            to_object: patient
            via_relation: patient_has_variant
    """

    mapping = yaml.safe_load(yaml_mapping)

    logging.debug("Load data...")
    csv_file = "tests/" + directory_name + "/data.csv"
    table = pd.read_csv(csv_file)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, mapping, affix="none")


if __name__ == "__main__":
    test_transformer_user()



def test_affix_separator():
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "affix_separator"

    expected_nodes = [('patient___A', 'patient', {}),
                    ('patient___B', 'patient', {}),
                    ('patient___C', 'patient', {}),
                    ('publication___publicationA', 'publication', {}),
                    ('publication___publicationB', 'publication', {}),
                    ('publication___publicationC', 'publication', {}),
                    ('variant___0', 'variant', {}),
                    ('variant___1', 'variant', {}),
                    ('variant___2', 'variant', {}),
                    ]

    expected_edges = [('', 'variant___0', 'patient___A', 'patient_has_variant', {}),
                    ('', 'variant___0', 'publication___publicationA', 'publication_to_variant', {}),
                    ('', 'variant___1', 'patient___B', 'patient_has_variant', {}),
                    ('', 'variant___1', 'publication___publicationB', 'publication_to_variant', {}),
                    ('', 'variant___2', 'patient___C', 'patient_has_variant', {}),
                    ('', 'variant___2', 'publication___publicationC', 'publication_to_variant', {}),
                    ]

    data_mapping = {f"tests/{directory_name}/data.csv": f"tests/{directory_name}/mapping.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="prefix", type_affix_sep="___")

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_affix_separator()



import logging

import ontoweaver

def test_reconciliate():
    nodes = [
        ("1", "Source", {"p1":"z"}),
        ("1", "Source", {"p2":"y"}), # Simple duplicate.
        ("2", "Target", {}),
        ("2", "Target", {"p1":"x", "p2":"y"}),
        ("2", "Target", {"p2":"z"}),
    ]

    edges = [
        ("A", "1", "2", "Edge", {"q1":"i"}),
        ("B", "1", "2", "Edge", {"q2":"j"}),
        ("C", "2", "1", "Edge", {"q1":"i", "q2": "j"}),
    ]

    fnodes, fedges = ontoweaver.fusion.reconciliate(nodes, edges, reconciliate_sep=";")

    assert(len(fnodes) == 2)
    assert(len(fedges) == 2)

    for n in fnodes:
        assert("p1" in n[2]) # properties
        assert("p2" in n[2]) # properties
        assert(n[0] in "12") # id
        assert(n[1] in ["Source", "Target"]) # Label/type
        logging.info(n)

    for e in fedges:
        assert("q1" in e[4])
        assert("q2" in e[4])
        assert(e[3] == "Edge")
        for l in e[0].split(";"):
            assert(l in "ABC")
        assert(e[1] in "12")
        assert(e[2] in "12")
        logging.info(e)


if __name__ == "__main__":
    test_reconciliate()



""" Adapters managing iterable data.
"""
import logging
import threading

import pandas as pd

from typing import Optional
from itertools import chain
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from abc import ABCMeta as ABSTRACT, abstractmethod

from . import base
from . import transformer
from . import exceptions
from . import validate

logger = logging.getLogger("ontoweaver")

class IterativeAdapter(base.Adapter, metaclass = ABSTRACT):
    """Base class for implementing a Biocypher adapter that consumes iterative data."""

    def __init__(self,
                 subject_transformer: base.Transformer,
                 transformers: Iterable[base.Transformer],
                 metadata: Optional[dict] = None,
                 validator: Optional[validate.InputValidator] = None,
                 type_affix: Optional[base.TypeAffixes] = base.TypeAffixes.suffix,
                 type_affix_sep: Optional[str] = ":",
                 parallel_mapping: int = 0,
                 raise_errors = True,
                 ):
        """
        Instantiate the adapter.

        Args:
            subject_transformer (base.Transformer): The transformer that maps the subject node.
            transformers (Iterable[base.Transformer]): List of transformer instances that map the data frame to nodes and edges.
            metadata (Optional[dict]): Metadata to be added to all the nodes and edges.
            type_affix (Optional[TypeAffixes]): Where to add a type annotation to the labels (either TypeAffixes.prefix, TypeAffixes.suffix or TypeAffixes.none).
            type_affix_sep (Optional[str]): String used to separate a label from the type annotation (WARNING: double-check that your BioCypher config does not use the same character as a separator).
            parallel_mapping (int): Number of workers to use in parallel mapping. Defaults to 0 for sequential processing.
            raise_errors (bool): if True, will raise an exception when an error is encountered, else, will log the error and try to proceed.
        """
        super().__init__(raise_errors)

        self.validator = validator

        if type_affix not in base.TypeAffixes:
            self.error(f"`type_affix`={type_affix} is not one of the allowed values ({list(base.TypeAffixes)})", exception = exceptions.ConfigError)
        else:
            self.type_affix = type_affix

        self.type_affix_sep = type_affix_sep

        self.subject_transformer = subject_transformer
        assert self.subject_transformer
        self.transformers = transformers
        self.property_transformers = [] # populated at parsing in self.properties.
        self.metadata = metadata
        # logger.debug(self.target_element_properties)
        self.parallel_mapping = parallel_mapping

        self.non_viable_rows = set()


    # FIXME not used, maybe will come in handy?
    def source_type(self, row):
        """
        Accessor to the row type actually used by `run`.

        You may overload this function if you want
        to make the row type dependent on some column value.

        By default, just return the default row type defined in the constructor,
        without taking the row values into account.

        Args:
            row: The current row of the DataFrame.

        Returns:
            The row type.
        """
        return self.row_type


    def make_id(self, entry_type, entry_name):
        """
        LabelMaker a unique ID for the given cell consisting of the entry name and type,
        taking into account affix and separator configuration.

        Args:
            entry_type: The type of the entry.
            entry_name: The name of the entry.

        Returns:
            str: The created ID.

        Raises:
            ValueError: If the ID creation fails.
        """
        assert isinstance(entry_type, str), "Entry type is not a string"
        if not isinstance(entry_name, str):
            logger.warning(f"Identifier `{entry_name}` (of type `{entry_type}`) is not a string, I had to convert it explicitely, check that the related transformer yields a string.")
            entry_name = str(entry_name)

        if '[' in entry_name or ']' in entry_name:
            logger.warning(f"Identifier `{entry_name}` (of type `{entry_type}`) contains brackets. Maybe you should use a `split` transformer for this column?")

        if self.type_affix == base.TypeAffixes.prefix:
            idt = f'{entry_type}{self.type_affix_sep}{entry_name}'
        elif self.type_affix == base.TypeAffixes.suffix:
            idt = f'{entry_name}{self.type_affix_sep}{entry_type}'
        elif self.type_affix == base.TypeAffixes.none:
            idt = f'{entry_name}'

        if idt:
            logger.debug(f"\t\tFormatted ID `{idt}` for cell value `{entry_name}` of type: `{entry_type}`")
            return idt
        else:
            self.error(f"Failed to format ID for cell value: `{entry_name}` of type: `{entry_type}`", exception = exceptions.DeclarationError)


    def valid(self, val):
        """
        Checks if cell value is valid - not a `nan`.

        Args:
            val: The value to check.

        Returns:
            bool: True if the value is valid, False otherwise.
        """
        if pd.api.types.is_numeric_dtype(type(val)):
            if (math.isnan(val) or val == float("nan")):
                return False
        elif str(val) == "nan":  # Conversion from Pandas' `object` needs to be explicit.
            return False
        return True


    def properties(self, property_dict, row, i, edge_t, node_t, node = False):

        """
        Extract properties of each property category for the given node type.
        If no properties are found, return an empty dictionary.

        Args:
            property_dict: Dictionary of transformer => property name
            row: The current row of the DataFrame.
            i: The index of the current row.
            edge_t: The type of the edge of the current transformer.
            node_t: The type of the node of the current transformer.
            node: True if the object created is a node, False otherwise.

        Returns:
            dict: Extracted properties.
        """
        properties = {}

        if property_dict:

            for prop_transformer, property_name in property_dict.items():
                properties[property_name] = []

                for property_value, none_node, none_edge, none_reverse_relation \
                    in prop_transformer(row, i):

                    if property_value:
                        if isinstance(property_value, list):
                            properties[property_name] += str(property_value)
                        else:
                            properties[property_name].append(str(property_value))
                        # properties[property_name] = str(property_value).replace("'", "`") # FIXME double-check why the old code needed this.
                        logger.debug(
                            f"                 {prop_transformer}" \
                            f" to property `{property_name}` with value" \
                            f" `{properties[property_name]}`." )
                    else:
                        self.error(
                            f"Failed to extract valid property with" \
                            f" {prop_transformer.__repr__()} for {i}th row.",
                            indent=2,
                            exception = exceptions.TransformerDataError )

        # Collapse any list with less than one item to a string.
        # to_delete = []
        for k,v in properties.items():
            # if len(v) == 0:
            #     to_delete.append(k)
            # el
            if len(v) == 1:
                properties[k] = v[0]

        # We don't delete empty properties, because BioCypher would complain.
        # for k in to_delete:
            # del properties[k]

        # If the metadata dictionary is not empty, add the metadata to the property dictionary.
        if self.metadata:
            if node:
                elem = node_t
            else:
                elem = edge_t
            if elem.__name__ in self.metadata:
                for key, value in self.metadata[elem.__name__].items():
                    properties[key] = value

        return properties


    def make_node(self, node_t, id, properties):
        """
        LabelMaker nodes of a certain type.

        Args:
            node_t: The type of the node.
            id: The ID of the node.
            properties: The properties of the node.

        Returns:
            The created node.
        """
        return node_t(id=id, properties=properties)


    def make_edge(self, edge_t, id_target, id_source, properties):
        """
        LabelMaker edges of a certain type.

        Args:
            edge_t: The type of the edge.
            id_target: The ID of the target node.
            id_source: The ID of the source node.
            properties: The properties of the edge.

        Returns:
            The created edge.
        """
        default_id = f"({id_source})--[{edge_t.__name__}]->({id_target})"
        return edge_t(id = default_id, id_source=id_source, id_target=id_target, properties=properties)

    # ==========================
    # Helper functions for run
    # ==========================

    def _make_default_source_node_id(self, row, i, local_nodes, local_errors):
        """
        Helper function to create the default source node id for each row.
        Referred to as default because of possibility
        of it changing type with `from_subject` attribute in transformers.
        """

        logger.debug("\tMake subject node...")
        for source_id, subject_edge_type, subject_node_type, _ in self.subject_transformer(row, i):
        # subject_generator_list = list(self.subject_transformer(row, i))
        # logger.debug(f"\t\tSubject items: {subject_generator_list}")
        # if len(subject_generator_list) > 1:
        #     local_errors.append(self.error(
        #         f"You cannot use transformer yielding multiple IDs for the subject. "
        #         f"Subject Transformer `{self.subject_transformer}` produced multiple IDs: "
        #         f"{subject_generator_list}, I'll take the first one and pretend this"
        #         f" did not happened.",
        #         indent=2, exception=exceptions.TransformerInterfaceError))

        # elif len(subject_generator_list) == 0:
        #     logger.debug("The subject transformer did not produce any valid ID,"
        #         " I'll try to skip this entry.")
        #     return None
        #
        # source_id, subject_edge_type, subject_node_type, _ = subject_generator_list[0]

            if self.subject_transformer.final_type:
                # If a final_type attribute is present in the transformer, use it as the source node type, instead
                # of the default type.
                subject_node_type = self.subject_transformer.final_type

            assert subject_node_type

            if source_id:
                source_node_id = self.make_id(subject_node_type.__name__, source_id)

                if source_node_id:
                    logger.debug(f"\t\tDeclared subject ID: {source_node_id}")
                    local_nodes.append(
                        self.make_node(
                            node_t = subject_node_type,
                            id = source_node_id,
                            # FIXME: Should we use the meta-way of accessing node properties as well?
                            # FIXME: This would require a refactoring of the transformer interfaces and tabular.run.
                            properties = self.properties(
                                self.subject_transformer.properties_of,
                                row,
                                i,
                                subject_edge_type,
                                subject_node_type,
                                node = True
                            )
                        )
                    )
                else:
                    local_errors.append(self.error(
                        f"Failed to declare subject ID for row #{i}: `{row}`.",
                        indent = 2,
                        exception = exceptions.DeclarationError ))

                # return source_node_id
                yield source_node_id

            else:
                local_errors.append(self.error(
                    f"No valid source node identifier from {self.subject_transformer} for {i}th row."
                    f" This row will be skipped.",
                    indent = 2,
                    exception = exceptions.TransformerDataError ))


    def _make_target_node_id(self, row, i, transformer, j, target_id, target_edge, target_node, local_nodes, local_errors):
        """
        Helper function to create the target node id for each target transformer.
        """

        if target_id and target_edge and target_node:

            if transformer.final_type:
                # If a final_type attribute is present in the transformer, use it as the target node type, instead
                # of the default type.
                target_node = transformer.final_type

            target_node_id = self.make_id(target_node.__name__, target_id)
            logger.debug(f"\t\tMake node {target_node_id}")
            local_nodes.append(self.make_node(node_t=target_node, id=target_node_id,
                                              # FIXME: Should we use the meta-way of accessing node properties as well?
                                              # FIXME: This would require a refactoring of the transformer interfaces and tabular.run.
                                              properties=self.properties(transformer.properties_of, row,
                                                                         i, target_edge, target_node, node=True)))

        else:
            local_errors.append(self.error(f"No valid target node identifier from {transformer}"
                                           f" for {i}th row.", indent=2, section="transformers",
                                           index=j, exception=exceptions.TransformerDataError))

            target_node_id = None

        return target_node_id

    def _make_alternative_source_node_id(self, row, i, transformer, j, target_node_id, target_edge, local_edges, local_errors):
        """
        Helper function to create an alternative source node id, in case the target transformer has a `from_subject` attribute.
        """

        found_valid_subject = False

        for t in self.transformers:
            if transformer.from_subject == t.target_type:
                found_valid_subject = True
                for s_id, s_edge, s_node, s_reverse_edge in t(row, i):
                    if s_id and s_edge and s_node:
                        if t.final_type:
                            s_node = t.final_type
                        subject_id = s_id
                        subject_node_id = self.make_id(t.target_type, subject_id)
                        logger.debug(
                            f"\t\tMake edge from {subject_node_id} toward {target_node_id}")
                        local_edges.append(
                            self.make_edge(edge_t=target_edge, id_source=subject_node_id,
                                           id_target=target_node_id,
                                           properties=self.properties(target_edge.fields(),
                                                                      row, i, s_edge, s_node)))

                    else:
                        local_errors.append(self.error(
                            f"No valid identifiers from {t} for {i}th row, when trying to change default subject type",
                            f"by {transformer} with `from_subject` attribute.",
                            indent=7, section="transformers", index=j,
                            exception=exceptions.TransformerDataError))

            else:
                # The transformer instance type does not match the type in the `from_subject` attribute.
                continue

        if not found_valid_subject:
            local_errors.append(self.error(f"\t\t\tInvalid subject declared from {transformer}."
                                           f" The subject you declared in the `from_subject` directive: `"
                                           f"{transformer.from_subject}` must not be the same as the default subject type.",
                                           exception=exceptions.ConfigError))


    def _no_element(self, local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes):
        if local_nodes == None or local_edges == None or local_errors == None \
            or local_rows == None or local_transformations == None or local_nb_nodes == None:
            logger.debug(f"There's a None in: {local_nodes}, {local_edges}, {local_errors}, {local_rows}, {local_transformations}, {local_nb_nodes}")
            return True
        return False


    def _run_all(self, process_row, nb_rows, nb_transformations, nb_nodes):
        """
        Perform the final logging after processing all rows.
        """

        if self.parallel_mapping > 0:
            logger.info(f"Processing dataframe in parallel. Number of workers set to: {self.parallel_mapping} ...")
            # Process the dataset in parallel using ThreadPoolExecutor
            with ThreadPoolExecutor() as executor:
                # Map the process_row function across the dataframe
                results = list(executor.map(process_row, self.iterate() ))

            # Append the results in a thread-safe manner after all rows have been processed
            i = 0
            for local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes in results:
                i += 1
                if self._no_element(local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes):
                    # logger.warning(f"Processing row {i} led to no viable element, I'll just skip it.")
                    self.non_viable_rows.add(i)
                    continue
                with self._nodes_lock:
                    self.nodes_append(local_nodes)
                with self._edges_lock:
                    self.edges_append(local_edges)
                with self._errors_lock:
                    self.errors += local_errors
                with self._row_lock:
                    nb_rows += local_rows
                with self._transformations_lock:
                    nb_transformations += local_transformations
                with self._local_nb_nodes_lock:
                    nb_nodes += local_nb_nodes

            assert len(self.nodes) > 0

        elif self.parallel_mapping == 0:
            logger.debug("Processing data sequentially...")
            for i, row in self.iterate():
                local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes = process_row((i, row))
                if self._no_element(local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes):
                    # logger.warning(f"Processing row {i} led to no viable element, I'll just skip it.")
                    self.non_viable_rows.add(i)
                    continue
                self.nodes_append(local_nodes)
                self.edges_append(local_edges)
                self.errors += local_errors
                nb_rows += local_rows
                nb_transformations += local_transformations
                nb_nodes += local_nb_nodes
                yield local_nodes, local_edges

        else:
            self.error(f"Invalid value for `parallel_mapping` ({self.parallel_mapping})."
                       f" Pass 0 for sequential processing, or the number of workers for parallel processing.", exception = exceptions.ConfigError)

        # Final logger
        error_count = {}
        for transformer in chain([self.subject_transformer], self.transformers, self.property_transformers):
            if transformer.output_validator:
                for msg,err in transformer.output_validator.messages.items():
                    desc = f"in {err['section']} with {transformer}: {msg}"
                    # FIXME do we want the number of validation errors or the number of bad cell values?
                    # error_count[desc] = error_count.get(desc, 0) + err['count']
                    error_count[desc] = error_count.get(desc, err['count'])
        for desc,count in error_count.items():
            logger.error(f"Recorded {count} times a validation error {desc}")

        if self.errors:
            logger.error(
                f"Recorded {len(self.errors)} errors while processing {nb_transformations} transformations with {1+len(self.transformers)} node transformers, producing {nb_nodes} nodes for {nb_rows} rows.")
            # logger.debug("\n".join(self.errors))
        else:
            logger.info(
                f"Performed {nb_transformations} transformations with {1+len(self.transformers)} node transformers, producing {nb_nodes} nodes for {nb_rows} rows.")

    # =============
    # Run function
    # =============

    def run(self):
        """Iterate through dataframe in parallel and map cell values according to YAML file, using a list of transformers."""

        logger.debug("Run...")
        # Thread-safe containers with their respective locks
        self._nodes = []
        self._edges = []
        self._errors = []
        self._nodes_lock = threading.Lock()
        self._edges_lock = threading.Lock()
        self._errors_lock = threading.Lock()
        self._row_lock = threading.Lock()
        self._transformations_lock = threading.Lock()
        self._local_nb_nodes_lock = threading.Lock()

        # Function to process a single row and collect operations
        def process_row(row_data):
            i, row = row_data
            local_nodes = []
            local_edges = []
            local_errors = []
            local_rows = 0
            local_transformations = 0 # Count of transformations for this row. Does not include the property transformers.
            local_nb_nodes = 0

            logger.debug(f"Process row {i}...")
            local_rows += 1

            # source_node_id = self._make_default_source_node_id(row, i, local_nodes, local_errors)
            # if not source_node_id:
            #     return None, None, None, None, None, None
            for source_node_id in self._make_default_source_node_id(row, i, local_nodes, local_errors):
                logger.debug(f"Got subject node id: `{source_node_id}`")
                if source_node_id:
                    local_nb_nodes += 1
                else:
                    return None, None, None, None, None, None

                # Loop over list of transformer instances and label_maker corresponding nodes and edges.
                # FIXME the transformer variable here shadows the transformer module.
                for j,transformer in enumerate(self.transformers):
                    local_transformations += 1
                    logger.debug(f"\tCalling the {j}th transformer: {transformer}...")
                    k = 0
                    for target_id, target_edge, target_node, reverse_relation in transformer(row, i):
                        logger.debug(f"\t\t{k}th element yielded by transformer")
                        k += 1
                        target_node_id = self._make_target_node_id(
                            row,
                            i,
                            transformer,
                            j,
                            target_id,
                            target_edge,
                            target_node,
                            local_nodes,
                            local_errors
                        )

                        #If no valid target node id was created, an error is logged in the `_make_target_node_id` function,
                        #and we move to the next iteration of the loop.
                        if target_node_id is None:
                            continue
                        else:
                            local_nb_nodes += 1

                            # If a `from_subject` attribute is present in the transformer, loop over the transformer
                            # list to find the transformer instance mapping to the correct type, and then label_maker new
                            # subject id.

                            # FIXME add hook functions to be overloaded.

                            # FIXME: Make from_subject reference a list of subjects instead of using the add_edge function.

                            if hasattr(transformer, "from_subject"):

                                self._make_alternative_source_node_id(
                                    row,
                                    i,
                                    transformer,
                                    j,
                                    target_node_id,
                                    target_edge,
                                    local_edges,
                                    local_errors
                                )

                            else: # no attribute `from_subject` in `transformer`
                                logger.debug(f"\t\tMake edge {target_edge.__name__} from {source_node_id} toward {target_node_id}")
                                local_edges.append(
                                    self.make_edge(
                                        edge_t=target_edge,
                                        id_target=target_node_id,
                                        id_source=source_node_id,
                                        properties=self.properties(
                                            target_edge.fields(),
                                            row,
                                            i,
                                            target_edge,
                                            target_node
                                        )
                                    )
                                )

                                if reverse_relation:
                                    logger.info(f"\t\t\tMake reverse edge {reverse_relation.__name__} from {target_node_id} to {source_node_id}")
                                    local_edges.append(
                                        self.make_edge(
                                            edge_t=reverse_relation,
                                            id_target=source_node_id,
                                            id_source=target_node_id,
                                            properties=self.properties(
                                                reverse_relation.fields(),
                                                row,
                                                i,
                                                reverse_relation,
                                                source_node_id.__class__
                                            )
                                        )
                                    )
            # assert hasattr(local_nodes, "__iter__")
            # assert hasattr(local_edges, "__iter__")
            return local_nodes, local_edges, local_errors, local_rows, local_transformations, local_nb_nodes
            # End of process_row local function

        nb_rows = 0
        nb_transformations = 0
        nb_nodes = 0

        if self.parallel_mapping > 0:
            logger.debug("\tParallel mapping...")
            self._run_all(process_row, nb_rows, nb_transformations, nb_nodes)
        else:
            logger.debug("\tSequential mapping...")
            for local_nodes,local_edges in self._run_all(process_row, nb_rows, nb_transformations, nb_nodes):
                yield local_nodes, local_edges

    def __call__(self):
        # FIXME If the run functions contains a generator nested under an `if`, the call function is note even called by the derivated class instance..."
        # if self.parallel_mapping > 0:
        #     logger.debug("Parallel mapping")
        #     self.run()
        #     return
        # else:
        #     logger.debug("Sequential mapping")
        for local_nodes, local_edges in self.run():
            yield local_nodes, local_edges

    @abstractmethod
    def iterate(self):
        """ The function that access the atomic item of its targeted iterable data structure.

            It is called by run() in a way that is equivalent to:

            .. code-block:: python

                for i,item in self.iterate():

            For example, for a table, it yield the row number and the row itself.

            Returns:
                The index of the item and the item.
        """
        raise NotImplementedError

    def __del__(self):
        if self.non_viable_rows:
            logger.warning( \
                f"Got {len(self.non_viable_rows)} rows that produced no valid element." \
                " I skipped them silently, but you may want to double check your mapping." \
                " This is normal if you used a match section that does not consider" \
                " all possible values. Run with the DEBUG log level to see the row numbers.")
            logger.debug(", ".join(str(i) for i in self.non_viable_rows))




def test_multiple_databases():
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "multiple_databases"

    expected_nodes = [
        ('0:variant', 'variant', {}),
        ('1:variant', 'variant', {}),
        ('2:variant', 'variant', {}),
        ('ENST00000380956:transcript', 'transcript', {}),
        ('ENST00000523534:transcript', 'transcript', {}),
        ('ENST00000651671:transcript', 'transcript', {}),
        ('IRF4:gene_hugo', 'gene_hugo', {}),
        ('NOTCH1:gene_hugo', 'gene_hugo', {}),
        ('NRG1:gene_hugo', 'gene_hugo', {}),
        ('Olaparib:drug', 'drug', {}),
        ('PM_1:publication', 'publication', {}),
        ('PM_2:publication', 'publication', {}),
        ('PM_3:publication', 'publication', {}),
        ('Palbociclib:drug', 'drug', {}),
        ('Tazemetostat:drug', 'drug', {}),
        ('non-oncogenic:disease', 'disease', {}),
        ('oncogenic:disease', 'disease', {}),
        ('patient_1:patient', 'patient', {}),
        ('patient_2:patient', 'patient', {}),
        ('patient_3:patient', 'patient', {}),
    ]

    expected_edges = [
        ('', '0:variant', 'NRG1:gene_hugo', 'variant_in_gene', {}),
        ('', '0:variant', 'Palbociclib:drug', 'variant_affected_by_drug', {}),
        ('', '0:variant', 'non-oncogenic:disease', 'variant_to_disease', {}),
        ('', '0:variant', 'patient_1:patient', 'patient_has_variant', {}),
        ('', '1:variant', 'IRF4:gene_hugo', 'variant_in_gene', {}),
        ('', '1:variant', 'Tazemetostat:drug', 'variant_affected_by_drug', {}),
        ('', '1:variant', 'non-oncogenic:disease', 'variant_to_disease', {}),
        ('', '1:variant', 'patient_2:patient', 'patient_has_variant', {}),
        ('', '2:variant', 'NOTCH1:gene_hugo', 'variant_in_gene', {}),
        ('', '2:variant', 'NRG1:gene_hugo', 'variant_in_gene', {}),
        ('', '2:variant', 'Olaparib:drug', 'variant_affected_by_drug', {}),
        ('', '2:variant', 'oncogenic:disease', 'variant_to_disease', {}),
        ('', '2:variant', 'patient_3:patient', 'patient_has_variant', {}),
        ('', 'IRF4:gene_hugo', 'ENST00000380956:transcript', 'transcript_to_gene_relationship', {}),
        ('', 'NOTCH1:gene_hugo', 'ENST00000651671:transcript', 'transcript_to_gene_relationship', {}),
        ('', 'NRG1:gene_hugo', 'ENST00000523534:transcript', 'transcript_to_gene_relationship', {}),
        ('', 'Olaparib:drug', 'PM_3:publication', 'treatment_has_citation', {}),
        ('', 'Palbociclib:drug', 'PM_1:publication', 'treatment_has_citation', {}),
        ('', 'Tazemetostat:drug', 'PM_2:publication', 'treatment_has_citation', {}),
    ]

    list_nodes = []
    list_edges = []

    data_mapping = {f"tests/{directory_name}/data_cgi_article.csv": f"tests/{directory_name}/cgi.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix", raise_errors=False)

    list_nodes += nodes
    list_edges += edges

    data_mapping = {f"tests/{directory_name}/data_oncokb_article.csv": f"tests/{directory_name}/oncokb.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix", raise_errors=False)

    list_nodes += nodes
    list_edges += edges

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(list_nodes), ontoweaver.ow2bc(list_edges), reconciliate_sep=",")

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_multiple_databases()



import os
import logging
import subprocess

import ontoweaver

class user_transformer(ontoweaver.base.Transformer):
    def __init__(self, target, properties_of, edge=None, columns=None, **kwargs):
        super().__init__(target, properties_of, edge, columns, **kwargs)

    def __call__(self, row, i):
        for key in self.columns:
            yield str(row[key])


def test_ontoweave():
    logging.debug(f"From: {os.getcwd()}")
    cmd="ontoweave --biocypher-config ./tests/simplest/biocypher_config.yaml --biocypher-schema ./tests/simplest/schema_config.yaml --type-affix suffix --type-affix-sep : --prop-sep ';' --debug ./tests/simplest/data.csv:./tests/simplest/mapping.yaml"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


def test_ontoweave_register():
    logging.debug(f"From: {os.getcwd()}")
    cmd="ontoweave --biocypher-config ./tests/simplest/biocypher_config.yaml --biocypher-schema ./tests/simplest/schema_config.yaml --type-affix suffix --type-affix-sep : --prop-sep ';' --debug ./tests/simplest/data.csv:./tests/simplest/mapping.yaml --register ./tests/test_ontoweave.py"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


def test_ontoweave_automap():
    logging.debug(f"From: {os.getcwd()}")
    cmd="ontoweave --biocypher-config ./tests/family_automap/biocypher_config_2_bioPathNet.yaml --biocypher-schema ./tests/family_automap/schema_config.yaml ./tests/family_automap/reasoned.ttl:automap --debug"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


def test_ontoweave_autoschema_min():
    logging.debug(f"From: {os.getcwd()}")
    d = "./tests/autoschema"
    xs = f"{d}/extended_schema.yaml"
    if os.path.isfile(xs):
        os.remove(xs)
    cmd = f"ontoweave --biocypher-config {d}/biocypher_config.yaml --auto-schema {xs} --biocypher-schema {d}/schema_min.yaml {d}/example.csv:{d}/mapping.yaml {d}/example.csv:{d}/mapping_with_props.yaml --debug"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


def test_ontoweave_autoschema_complete():
    logging.debug(f"From: {os.getcwd()}")
    d = "./tests/autoschema"
    xs = f"{d}/extended_schema.yaml"
    if os.path.isfile(xs):
        os.remove(xs)
    cmd = f"ontoweave --biocypher-config {d}/biocypher_config.yaml --auto-schema {xs} --biocypher-schema {d}/schema_complete.yaml {d}/example.csv:{d}/mapping.yaml {d}/example.csv:{d}/mapping_with_props.yaml --debug"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


def test_ontoweave_autoschema_half():
    logging.debug(f"From: {os.getcwd()}")
    d = "./tests/autoschema"
    xs = f"{d}/extended_schema.yaml"
    if os.path.isfile(xs):
        os.remove(xs)
    cmd = f"ontoweave --biocypher-config {d}/biocypher_config.yaml --auto-schema {xs} --biocypher-schema {d}/schema_half.yaml {d}/example.csv:{d}/mapping.yaml {d}/example.csv:{d}/mapping_with_props.yaml --debug"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


if __name__ == "__main__":
    test_ontoweave()



import ontoweaver
print("Module imported from:",ontoweaver.__file__)

import logging
logger = logging.getLogger("ontoweaver")

def test_congregate():
    nodes = [
        ("Source:1", "Source", {"p1":"z"}),
        ("Source:1", "Source", {"p2":"y"}), # Simple duplicate.
        ("Target:2", "Target", {"p1":"x", "p2":"y"}),
    ]

    edges = [
        ("Link:0", "Source:1", "Target:2", "Link", {}),
        ("Link:0", "Source:1", "Target:2", "Link", {}) # Complete duplicate
    ]
    logger.debug(nodes)

    on_ID = ontoweaver.serialize.ID()

    congregate_nodes = ontoweaver.congregate.Nodes(on_ID)
    print("nodes:",nodes)
    for n in congregate_nodes(nodes):
        pass

    assert nodes, "There is no node"
    logger.debug(congregate_nodes.duplicates)
    print("duplicates:",congregate_nodes.duplicates)
    assert len(congregate_nodes.duplicates) > 0, "There is no fused node"
    assert len(congregate_nodes.duplicates) == 2, "Not the correct number of duplicated nodes"

    congregate_edges = ontoweaver.congregate.Edges(on_ID)
    for n in congregate_edges(edges):
        pass

    assert edges, "There is no edge"
    assert len(congregate_edges.duplicates) == 1, "Not the correct number of duplicated nodes"


    on_Everything = ontoweaver.serialize.All()

    congregate_nodes = ontoweaver.congregate.Nodes(on_Everything)
    for n in congregate_nodes(nodes):
        pass
    assert nodes, "There is no node"
    assert len(congregate_nodes.duplicates) > 0, "There is no fused node"
    assert len(congregate_nodes.duplicates) == 3, "Not the correct number of duplicated nodes"

    congregate_edges = ontoweaver.congregate.Edges(on_Everything)
    for n in congregate_edges(edges):
        pass
    assert edges, "There is no edge"
    assert len(congregate_edges.duplicates) == 1, "Not the correct number of duplicated nodes"


if __name__ == "__main__":
    logger.setLevel("DEBUG")
    # logger.info("START")
    logging.getLogger("ontoweaver").setLevel("DEBUG")
    print("START")
    from biocypher._logger import get_logger as biocypher_logger
    biocypher_logger("biocypher").setLevel("DEBUG")
    test_congregate()



#!/usr/bin/env -S uv run --script
# /// script
# dependencies = [
#    "biocypher<1.0.0,>=0.11.0",
#    "pooch<2.0.0,>=1.7.0",
#    "pandas<3.0.0,>=2.3.1",
#    "numpy<3.0.0,>=2.2.4",
#    "owlready2<1.0,>=0.49",
#    "jsonargparse<5.0,>=4.39",
#    "xdg-base-dirs<7.0.0,>=6.0.2",
#    "pandera[io]<1.0.0,>=0.27.0",
#    "alive-progress<4.0,>=3.2",
#    "fsspec<2026.0.0,>=2025.10.0",
#    "natsort>5.0.0",
#    "lxml>6.0.0",
#    "jmespath>=1.0.1",
# ]
# ///

'''
Pre-processing of ontologies in order to be compatible with the requirements of Biocypher.

.. code:: sh

    python3 -m preprocess_ontology <ontology_to_be_transformed>[.owl] > <new_ontology>

This creates:

    - a new Biocypher compatible ontology, in <ontology_directory> named
      bc_<ontology_to_be_transformed>.owl
    - a 'bc_classes_mapping.json' file containing the mapping
      between classes from <ontology_to_be_transformed>.owl and classes from
      bc_<ontology_to_be_transformed>.owl

Each element of bc_classes_mapping.json file is a json input with the folloging information:

.. code::

    <initial_classe_IRI in ontology>: {
        "labels": [
            <list of the labels in the initial ontology>
            ],
        "bc_label": "<label in bc_ontology>"
    }

'''
import os
import re
import sys
import owlready2 as owl
import types
import json
import copy
import io
import rdflib
import logging

from itertools import chain

from rdflib import URIRef, Literal, Namespace
from rdflib.namespace import RDF, RDFS, OWL

logger = logging.getLogger("owl_to_biocypher")


chars_to_be_removed = [' ', "%", "-"]


def remove_characters(s, list_c):
    for c in list_c:
        s = s.replace(c, "")
    return s


def replace_underscore(s):
    p = s.find("_")
    while p>=0:
        c = s[p+1]
        s2 = s[:p]+c.upper()
        if len(s)>=p+1:
            s2=s2+s[p+2:]
        s = s2
        p = s.find("_")
    return s


def get_label_from_iri(iri):
    if iri.rfind("#")>0:
        return iri[iri.rfind("#")+1:]
    elif iri.rfind("/")>0:
        return iri[iri.rfind("/")+1:]
    else:
        return iri


def translate_labels(ontology_file, json_f=None, output_format="rdfxml"):
    ontology_path = os.path.abspath(ontology_file)
    iri_path = "".join(['file://', ontology_path])
    onto = owl.get_ontology(iri_path).load()

    translation_dict = {}

    with onto:
        thing_label = owl.Thing.label
        #print("\nThing label =", thing_label, type(thing_label))
        if not thing_label:
            thing_label = ["Thing"]
            owl.Thing.label = thing_label
        #print("Thing new label =", owl.Thing.label)

        for c in chain(onto.classes(), onto.object_properties()):
            # print(c, file=sys.stderr)
            # print(dir(c), file=sys.stderr)
            old_labels = c.label

            # labels = [l for l in c.label]
            # for l in labels:
            #     new_label = remove_characters(l, chars_to_be_removed)#.lower()
            #     new_label = replace_underscore(new_label)
            #     new_label = new_label.capitalize()
            #     c.label.append(new_label)

            # new_name = remove_characters(c.iri, chars_to_be_removed)
            # new_name = replace_underscore(new_name)
            # new_name = new_name.capitalize()
            if len(c.label)>0:
                new_label = c.label[0]
            else:
                new_label = get_label_from_iri(c.iri)#.lower()
            new_label = remove_characters(new_label, chars_to_be_removed)#.lower()
            new_label = replace_underscore(new_label)
            new_label = new_label[0].lower()+new_label[1:]
            c.label = []
            c.label.append(new_label)

            # If there is no obvious translation.
            if len(old_labels) == 0:
                logger.debug(f"There was no label for {new_label}, I will add it to the ontology, but not to the restoration for {c.iri}.")
            elif len(old_labels) == 1 and old_labels[0] == new_label:
                logger.debug(f"Label {new_label} is already compatible, I will not put it in the restoration for {c.iri}.")
            else:
                logger.info(f"Replace original {old_labels} by BioCypherized {new_label} for {c.iri}")
                # Save the translation.
                translation_dict[c.iri] = {
                    #"origin_class": c.iri,
                    "origin_labels": old_labels,
                    #"biocypher_class": new_name,
                    "biocypher_label": new_label
                }
            # parents = c.is_a
            # if parents == [owl.Thing]:
            #     with onto:
            #         bc_root_class = types.new_class("BcRootClass", (owl.Thing,))
            #         bc_root_class.label.append("BcRootClass")
            #         c.is_a = [bc_root_class]
            #print(c)

    if json_f:
        if len(translation_dict) == 0:
            logger.warning("There was no label translation to do on the input ontology, the restoration file will be a an empty mapping. But you still need to restore the ontology to remove the BioCypherRoot.")
        with open(json_f, 'w') as fp:
            json.dump(translation_dict, fp, indent=4)

    # Save ontology file in a buffer.
    by_io = io.BytesIO()
    onto.save(by_io, output_format)
    by_str = by_io.getvalue()
    text = by_str.decode("UTF-8")
    return text


def harden_labels(text):
    # Replace "<label>…</label>" by "<rdfs:label>…</rdfs:label>"
    # Because Biocypher needs rdfs:label, or else it does not found any class.
    # Because OwlReady2 does not allow label with prefixes,
    text = text.replace("<label", "<rdfs:label")
    text = text.replace("</label>", "</rdfs:label>")

    return text


def add_root(owl_txt, input_format, root_name = "BioCypherRoot"):
    # Separated function, because it's using rdflib instead of owlredy2,
    # because it's actually easier to manipulate low-level stuff.

    if input_format == "rdfxml":
        input_format = "xml"

    rdf_io = io.StringIO(owl_txt)
    g = rdflib.Graph()
    g.parse(rdf_io, format = input_format)

    bcns = Namespace("https://biocypher.org/biocypher#")
    g.bind("biocypher", bcns)

    namespaces = {}
    for key,ns in g.namespaces():
        namespaces[key] = Namespace(ns)

    # Add a BioCypherRoot on top of owl:Thing and owl:topObjectProperty
    uri_root = URIRef(namespaces["biocypher"][root_name])
    g.add( (
            uri_root,
            RDFS.label,
            Literal(root_name)
        ) )

    g.add( (
            OWL.Thing,
            RDFS.subClassOf,
            uri_root,
        ) )

    g.add( (
            OWL.topObjectProperty,
            RDFS.subPropertyOf,
            uri_root,
        ) )

    # Add owl:topObjectProperty as an ancestor of all owl:ObjectProperty in the default namespace
    g.add( (
            OWL.topObjectProperty,
            RDFS.label,
            Literal("topObjectProperty")
        ) )

    # For all ObjectProperty
    for s,p,o in g.triples((None, RDF.type, OWL.ObjectProperty)):
        # If not already inheriting from something.
        if (s,RDFS.subPropertyOf,None) not in g:
            # Makes it inheriting from topObjectProperty
            g.add( (
                    s,
                    RDFS.subPropertyOf,
                    OWL.topObjectProperty,
                ) )

    # Make Entity a Thing
    g.add( (
            URIRef(namespaces[""]["Entity"]),
            RDFS.subClassOf,
            OWL.Thing
        ) )

    return g


def harden_owl(ontology_file, json_f, output_format = "rdfxml"):
    # rdfxml (owlready2) / xml (rdflib) is our pivot format
    # to pass from one lib to the other.
    owl_txt = translate_labels(ontology_file, json_f, output_format)
    owl_txt = harden_labels(owl_txt)
    graph = add_root(owl_txt, input_format = "xml")
    return graph


if __name__ == "__main__":
    import sys
    import argparse
    appname = os.path.splitext(os.path.basename(sys.argv[0]))[0]

    p = argparse.ArgumentParser(
        description="Pre-process (i.e. downgrade) an OWL ontology to make it compatible with the (harsh) requirements of BioCypher.",
        epilog=f"Example usage: {appname} my-onto.owl --json my-restoration-.json > my-biocypherized-onto.owl")

    p.add_argument("ontology_file")

    p.add_argument("-j", "--json",
        help="a JSON file in which to save the computed mapping, for further reference or reconstruction (default: None)",
        metavar="JSON_FILE", default=None)

    rdflib_formats = ["xml", "n3", "turtle", "nt", "pretty-xml", "trix", "trig", "nquads", "json-ld", "hext"]
    owlready_formats = ["rdfxml","ntriples"]
    p.add_argument("-f", "--output-format",
        help="the format in which to write the ontology (default: turtle)",
        choices=rdflib_formats, default="turtle", metavar="FORMAT")

    p.add_argument("-l", "--log-level", default="WARNING",
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help="Configure the log level. [default: WARNING]")

    asked = p.parse_args()

    logging.basicConfig()
    logger.setLevel(asked.log_level)

    rdf_graph = harden_owl(asked.ontology_file, json_f = asked.json, output_format = "rdfxml")

    sys.stdout.write(rdf_graph.serialize(format = asked.output_format))



import logging
import yaml
import io
import pandas as pd

import ontoweaver

def test_transformer_nested():

    logging.debug("Load data...")

    # Do not add newlines or spaces here
    # or else the parsing will be wrong.
    data = """,name,aspects
0,A,{"is":"a"}
1,B,{"is":"b"}
2,C,{"is":"c"}
3,D,{"is":"d"}
"""
    csv = io.StringIO(data)
    table = pd.read_csv(csv)

    logging.debug("Load mappings...")

    mapping = """
row:
    map:
        column: name
        to_subject: Subject
transformers:
    - nested:
        keys:
            - aspects
            - is
        to_object: Aspect
        via_relation: is
    """

    ymap = yaml.full_load(mapping)

    logging.debug(ymap)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, ymap, affix="none")

    assert(len(nodes)==8)
    assert(len(edges)==4)


if __name__ == "__main__":
    test_transformer_nested()




import logging
import yaml
import io
import pandas as pd

import ontoweaver

def test_transformer_empty_values():

    logging.debug("Load data...")

    # Do not add newlines or spaces here
    # or else the parsing will be wrong.
    data = """,name,genre,is_child_of
0,father_1,Male,
1,child _nb_1_of_0,Female,father_1
2,child _nb_2_of_0,Female,father_1
3,child _nb_3_of_0,Male,father_1
4,father_2,Male,
5,father_3,Male,
6,child _nb_1_of_2,Male,father_3
7,child _nb_2_of_2,Male,father_3"""
    csv = io.StringIO(data)
    table = pd.read_csv(csv)

    logging.debug("Load mappings...")

    mapping = """
row:
    map:
        id_from_column: name
        match_type_from_column: genre
        match:
            - Male:
                to_subject: Male
            - Female:
                to_subject: Female
transformers:
    - map:
        column: is_child_of
        to_object: Male
        via_relation: is_child_of
    """

    ymap = yaml.full_load(mapping)

    logging.debug(ymap)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, ymap, affix="none")

    assert(len(nodes)==13)
    assert(len(edges)==5)


if __name__ == "__main__":
    test_transformer_empty_values()





def test_ontology_subtypes():
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "ontology_subtypes"

    expected_nodes = [
        ('0:variant', 'variant', {}),
        ('1:variant', 'variant', {}),
        ('2:variant', 'variant', {}),
        ('3:variant', 'variant', {}),
        ('fda1:fda_evidence_level', 'fda_evidence_level', {}),
        ('fda2:fda_evidence_level', 'fda_evidence_level', {}),
        ('fda3:fda_evidence_level', 'fda_evidence_level', {}),
        ('oncokb1:oncokb_evidence_level', 'oncokb_evidence_level', {}),
        ('oncokb2:oncokb_evidence_level', 'oncokb_evidence_level', {}),
        ('oncokb3:oncokb_evidence_level', 'oncokb_evidence_level', {}),
        ('patient1:patient', 'patient', {}),
        ('patient2:patient', 'patient', {}),
        ('patient3:patient', 'patient', {}),
        ('patient4:patient', 'patient', {}),
    ]

    expected_edges = [
        ('', '0:variant', 'fda1:fda_evidence_level', 'variant_to_evidence', {}),
        ('', '0:variant', 'oncokb1:oncokb_evidence_level', 'variant_to_evidence', {}),
        ('', '0:variant', 'patient1:patient', 'patient_has_variant', {}),
        ('', '1:variant', 'fda2:fda_evidence_level', 'variant_to_evidence', {}),
        ('', '1:variant', 'oncokb2:oncokb_evidence_level', 'variant_to_evidence', {}),
        ('', '1:variant', 'patient2:patient', 'patient_has_variant', {}),
        ('', '2:variant', 'fda3:fda_evidence_level', 'variant_to_evidence', {}),
        ('', '2:variant', 'oncokb3:oncokb_evidence_level', 'variant_to_evidence', {}),
        ('', '2:variant', 'patient3:patient', 'patient_has_variant', {}),
        ('', '3:variant', 'fda3:fda_evidence_level', 'variant_to_evidence', {}),
        ('', '3:variant', 'oncokb3:oncokb_evidence_level', 'variant_to_evidence', {}),
        ('', '3:variant', 'patient4:patient', 'patient_has_variant', {}),
    ]

    data_mapping = {f"tests/{directory_name}/data.csv": f"tests/{directory_name}/mapping.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix")

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges), reconciliate_sep=",")

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_ontology_subtypes()



import logging
import yaml
import io
import pandas as pd

import ontoweaver

def test_transformer_string():

    logging.debug("Load data...")

    # Do not add newlines or spaces here
    # or else the parsing will be wrong.
    data = """Patient,Variant,Source
P1,V1-1,S0
P1,V1-2,S1
P2,V2-1,S2
P2,V2-2,S3"""
    csv = io.StringIO(data)
    table = pd.read_csv(csv)

    logging.debug("Load mappings...")

    mapping = """
    row:
        map:
            columns:
                - Variant
            to_subject: variant
    transformers:
        - map:
            columns:
                - Patient
            to_object: patient
            via_relation: patient_has_variant
        - map:
            columns:
                - Source
            to_properties:
                - source
            for_objects:
                - patient
                - variant
        - string:
            value: "Whatever it is"
            to_properties:
                - something
            for_objects:
                - patient
                - variant
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="none")

    for node in nodes:
        assert(node.as_tuple()[2]["something"] == "Whatever it is")

if __name__ == "__main__":
    test_transformer_string()





#!/usr/bin/env -S uv run --script
# /// script
# dependencies = [
#    "biocypher<1.0.0,>=0.11.0",
#    "pooch<2.0.0,>=1.7.0",
#    "pandas<3.0.0,>=2.3.1",
#    "numpy<3.0.0,>=2.2.4",
#    "owlready2<1.0,>=0.49",
#    "jsonargparse<5.0,>=4.39",
#    "xdg-base-dirs<7.0.0,>=6.0.2",
#    "pandera[io]<1.0.0,>=0.27.0",
#    "alive-progress<4.0,>=3.2",
#    "fsspec<2026.0.0,>=2025.10.0",
#    "natsort>5.0.0",
#    "lxml>6.0.0",
#    "jmespath>=1.0.1",
# ]
# ///

""" The module holding utility functions for the main command line interface.
"""

import os
import sys
import yaml
import logging
import natsort
import pathlib
import platform
import xdg_base_dirs as xdg
import importlib
import networkx

import ontoweaver

# Additional error codes not handled by ontoweaver.exceptions.* classes.
error_codes = {
    "NetworkXError"   : 129,
    "Exception"       : 255,
}

def call_with_error_handling(func, *args, **kwargs):
    """Call the given function with the given argument,
    while system-exiting with the appropriate message and error code.
    """
    if "debug" in kwargs:
        debug = kwargs.pop("debug")

    if "section" in kwargs:
        sec = kwargs.pop("section")
        section = f" during {sec}"
    else:
        section = ""

    if debug:
        # In debug mode, let the exception ends the process and display the stack.
        return func(*args, **kwargs)
    else:
        # In non-debug mode, just display error messages and exit with error code.
        try:
            return func(*args, **kwargs)
        except ontoweaver.exceptions.OntoWeaverError as e:
            etype = type(e).__name__
            logging.error(f"{etype}{section} (in `ontoweave.{func.__name__}`)")
            edoc = type(e).__doc__
            if edoc:
                logging.error(edoc)
            logging.error(e)
            sys.exit(type(e).code)
        except networkx.exception.NetworkXError as e:
            logging.error(f"{type(e).__name__}{section} (in `ontoweave.{func.__name__}`).")
            logging.error(e)
            # probably "type not in the digraph"
            logging.error("Double check that you use the rdfs:label for this type in your schema, and not the IRI anchor, or look for any typo.")
            sys.exit(error_codes["NetworkXError"])
        except Exception as e:
            logging.error(f"UNKNOWN ERROR{section} (in `ontoweave.{func.__name__}`).")
            logging.error(e)
            sys.exit(error_codes["Exception"])


def check_file(filename):
    """Exit if the given filename does not exists or is not readable."""
    if not os.path.isfile(filename):
        logging.error(f"File `{filename}` not found.")
        # FIXME raise without exit if asked.debug
        sys.exit(ontoweaver.exceptions.FileError.code)

    if not os.access(filename, os.R_OK):
        logging.error(f"Cannot access file `{filename}`.")
        # FIXME raise without exit if asked.debug
        sys.exit(ontoweaver.exceptions.FileAccessError)


def config_directories(appname = "ontoweave"):
    """Yield standard configuration directories (as defined by XDG under MocOS/Unix)."""
    os = platform.system()
    logging.debug(f"Detected OS: {os}")

    #NOTE: All matched config files will be parsed and applied in the given order.
    if os == "Windows":
        if "APPDATA" in os.environ:
            yield pathlib.Path(os.environ["APPDATA"])
        yield pathlib.Path("~")/pathlib.Path("AppData")/pathlib.Path("Roaming")/pathlib.Path(appname)

    elif os == "Java" or os == "":
        logging.warning(f"I don't know where to search for configuration files on platform `{os}`, I'll only search in current directory")

    else: # Probably an Unix flavor (Darwin, Linux, Solaris, IRIX, etc.)
        # XDG will return the default defined in the specification
        # if the env variables are not set, so hardcoded defaults
        # like "~/.config/" should not be necessary.
        for p in xdg.xdg_config_dirs():
            yield p/appname
        yield xdg.xdg_config_home()/appname

    yield pathlib.Path("../src/ontoweaver")


def config_paths(appname = "ontoweave"):
    """Yield any path named <appname>.yaml in standard configuration directories."""
    dirs = config_directories()
    for p in dirs:
        yield str(p / (appname+".yaml"))


def main():
    import jsonargparse
    import argparse
    import subprocess
    import inspect

    appname = os.path.splitext(os.path.basename(sys.argv[0]))[0]

    logging.basicConfig()
    # logger = logging.getLogger(appname)
    logger = logging.getLogger("ontoweaver")

    config_files = list(config_paths(appname))

    do = jsonargparse.ArgumentParser(
        description = "A command line tool to run OntoWeaver mapping adapters on a set of tabular data, and call the created BioCypher export scripts.",
        epilog = f"Example usage:\n  {appname} table.csv:mapping.yaml\n  {appname} --biocypher-config config.yaml --auto-schema xschema.yaml table1.csv:mapping1.yaml table2.tsv:mapping1.yaml table3-part*.parquet:mapping2.yaml --import-script-run",
        default_config_files = config_files,
        env_prefix="ONTOWEAVE",
        default_env = True, # By default jsonargparse does not check environment, so one enables it explicitly.
        formatter_class = argparse.RawTextHelpFormatter,
        logger = logger, # FIXME jsonargparse seems to override any later config of logging, this seems like a bug.
    )

    do.add_argument("mapping", metavar="FILE:MAPPING", nargs="+",
        help=f"Run the given YAML MAPPING to extract data from the tabular FILE (usually a CSV). Several mappings can be passed to {appname}. You may also use the same mapping on different data files. If set to `STDIN`, will read the list of mappings from standard input.")

    do.add_argument("-c", "--config", metavar="FILE", action=jsonargparse.ActionConfigFile,
        help=f"The {appname} configuration file, which can host the same arguments than the command line tool. [default: {' or '.join(config_files)}]")

    do.add_argument("-C", "--biocypher-config", metavar="FILE", default="biocypher_config.yaml",
        help="The BioCypher config file (the one managing ontologies and the output backend). [default: %(default)s]")

    do.add_argument("-s", "--biocypher-schema", metavar="FILE", default="schema.yaml",
        help="The BioCypher schema file (the one managing node and edge types). [default: %(default)s]")

    do.add_argument("-m", "--auto-schema", metavar="FILE", default=None,
        help="Automatically generate a BioCypher schema from the mapping(s)," \
             " and save it into FILE." \
             " If an existing schema is passed with --biocypher-schema," \
             " then it is extended and saved into FILE." \
             " Raises an error if FILE already exists," \
             " unless --auto-schema-overwrite is passed. [default: %(default)s]")

    do.add_argument("-W", "--auto-schema-overwrite", action="store_true",
        help="Overwrite the schema generated with --auto-schema" \
             " if it exists [default: do not overwrite the auto schema]")

    do.add_argument("-p", "--parallel", metavar="NB_CORES", default="0",
        help=f"Number of processor cores to use when processing with multi-threading. `0` means a sequential processing (no parallelization, the default). Use 'auto' to let {appname} do its best to use a good number. [default: %(default)s]")

    do.add_argument("-i", "--import-script-run", action="store_true",
        help=f"If passed {appname} will call the import scripts created by Biocypher for you.")

    do.add_argument("-R", "--sort", metavar="KEY", default = "none",
        choices = ["none", "ascend"],
        help="Sort nodes and edges on their IDs. [default: %(default)s]")

    do.add_argument("-r", "--register", metavar="PYTHON_MODULE", nargs="*", default=[],
        help="Register all transformers available in the given module.")

    do.add_argument("-S", "--prop-sep", metavar="CHARACTER", default = ";",
        help="The character used to separate property values fused in the same property at the reconciliation step. [default: %(default)s]")

    do.add_argument("-a", "--type-affix", choices=["suffix","prefix","none"], default="none",
        help="Where to add the type string to the ID label. [default: %(default)s]")

    do.add_argument("-A", "--type-affix-sep", metavar="CHARACTER", default=":",
        help="Character used to separate the label from the type affix. [default: %(default)s]")

    do.add_argument("-E", "--pass-errors", action="store_true",
        help="When an error occurs, log is, and then try to continue processing. If not passed, the default behavior is to raise errors immediatly and stop execution.")

    do.add_argument("-l", "--log-level", default="WARNING",
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help="Configure the log level. [default: %(default)s]")

    do.add_argument("-v", "--validate-only", action="store_true",
                    help="Only validate the given input data, do not apply the mapping.")

    do.add_argument("-V", "--validate-output", action="store_true",
                    help="Validate the output data against the mapping rules.")

    do.add_argument("-e", "--pandas-sep", metavar="CHARACTER",
        help="Character used by the Pandas module to separate values in the input file database. [default: guessed from the extension]")

    do.add_argument("-D", "--debug", action="store_true",
        help="Run in debug mode. implies `--log-level DEBUG`, disables `--pass-errors`. NOTE: this will disable explicit return codes and show the call stack.")

    asked = do.parse_args()

    if asked.debug:
        # if asked.log_level != "DEBUG":
        #     logger.setLevel("DEBUG")
        #     logger.warning(f"You asked for --debug but set --log-level={asked.log_level}, I will ignore that and set --log-level=DEBUG")
        if asked.pass_errors:
            logger.warning("You asked for --debug but passed --pass-errors, I will ignore that.")
        # asked.log_level = "DEBUG"
        asked.pass_errors = False

    logger.setLevel(asked.log_level)
    logging.getLogger("ontoweaver").setLevel(asked.log_level)

    logger.info("OntoWeave parameters:")

    logger.info(f"    config files: {config_files}")
    logger.info(f"    config: `{asked.biocypher_config}`")
    logger.info(f"    schema: `{asked.biocypher_schema}`")
    logger.info(f"    auto-schema: `{asked.auto_schema}`")
    logger.info(f"    parallel: `{asked.parallel}`")
    logger.info(f"    prop-sep: `{asked.prop_sep}`")
    logger.info(f"    type-affix: `{asked.type_affix}`")
    logger.info(f"    type-affix-sep: `{asked.type_affix_sep}`")
    logger.info(f"    pandas-sep: `{asked.pandas_sep}`")
    logger.info(f"    import-script-run: `{asked.import_script_run}`")
    logger.info(f"    sort: `{asked.sort}`")
    logger.info(f"    register: `{asked.register}`")
    logger.info(f"    debug: `{asked.debug}`")
    logger.info(f"    log-level: `{asked.log_level}`")
    logger.info(f"    pass-errors: `{asked.pass_errors}`")
    logger.info(f"    validate-only: `{asked.validate_only}`")
    logger.info(f"    validate-output: `{asked.validate_output}`")

    logger.info(f"    asked mappings: `{asked.mapping}`")
    asked_mapping = []

    if asked.mapping == ["STDIN"]:
        while True:
            try:
                item = sys.stdin.readline()
            except UnicodeDecodeError:
                continue
            except KeyboardInterrupt:
                break
            if not item:
                break
            asked_mapping.append(item)
    else:
        asked_mapping = asked.mapping

    logger.info("    parsed mappings:")
    mappings = {}
    for data_map in asked_mapping:
        if ":" not in data_map:
            msg = f"Cannot parse the DATA:MAPPING `{data_map}`, I cannot find the colon character."
            logger.error(msg)
            sys.exit(ontoweaver.exceptions.ConfigError.code)
        data,map = data_map.split(":")
        mappings[data] = map
        logger.info(f"    `{data}` => `{map}`")

    if asked.parallel == "auto":
        parallel = min(32, (os.process_cpu_count() or 1) + 4)
    else:
        parallel = int(asked.parallel)
    logger.info(f"    parallel: `{asked.parallel}`")

    # Late import to avoid useless Biocypher's logs when asking for --help.
    from biocypher._logger import get_logger as biocypher_logger
    biocypher_logger("biocypher").setLevel(asked.log_level)

    # Validate the input data if asked.
    if asked.validate_only:
        logger.info("Validating input data frame...")
        call_with_error_handling(
            # function
            ontoweaver.validate_input_data,
            # its arguments
            filename_to_mapping = mappings,
            sep = asked.pandas_sep,
            raise_errors = not asked.pass_errors,
            # Error handling parameters
            debug = asked.debug,
            section = "validation of input data"
        )

    # Register all transformers existing in the given modules.
    ontoweaver.transformer.register_all( asked.register )

    # Double check file inputs and exit on according errors.
    check_file(asked.biocypher_config)

    # --auto-schema will generate this file instead of reading it.
    if not asked.auto_schema:
        check_file(asked.biocypher_schema)
    else:
        try:
            with open(asked.biocypher_schema) as fd:
                existing_schema = yaml.safe_load(fd)
        except Exception as e:
            logger.debug(e)
            msg = "There is no valid existing BioCypher schema," \
                  " I will create a completely new one."
            logger.info(msg)
            existing_schema = {}
        else:
            logger.info(f"Loaded the existing schema `{asked.biocypher_schema}`," \
                         " I will now extend it.")

        logger.info("Validating input data frame...")
        extended_schema_filename = call_with_error_handling(
            # function
            ontoweaver.autoschema,
            # its arguments
            asked.mapping,
            existing_schema,
            asked.auto_schema,
            asked.auto_schema_overwrite,
            not asked.pass_errors,
            # Error handling parameters
            debug = asked.debug,
            section = "validation of input data"
        )

        # Replace the schema filename with the autoschema one.
        logger.debug(f"Extend the original schema `{asked.biocypher_schema}` into `{extended_schema_filename}`")
        asked.biocypher_schema = extended_schema_filename

    for file_map in asked.mapping:
        data_file, map_file = file_map.split(":")
        if '*' not in data_file and '?' not in data_file:
            # Do not check if we are globbing several data files.
            check_file(data_file)

        if map_file != "automap":
            check_file(map_file)

    if asked.validate_output:
        validate_output = True
    else:
        validate_output = False

    kw = {}
    if asked.pandas_sep:
        kw = {"sep": asked.pandas_sep}

    if asked.sort == "none":
        sort_key = None
    elif asked.sort == "ascend":
        sort_key = natsort.natsort_keygen()
    else:
        msg = f"Unsupported sorting type `{asked.sort}`"
        logging.error(msg)
        sys.exit("ConfigError")

    logger.info(f"Running OntoWeaver...")
    import_file = call_with_error_handling(
        # function
        ontoweaver.weave,
        # its arguments
        asked.biocypher_config,
        asked.biocypher_schema,
        mappings,
        parallel_mapping = parallel,
        reconciliate_sep = asked.prop_sep,
        affix = asked.type_affix,
        type_affix_sep = asked.type_affix_sep,
        validate_output = validate_output,
        sort_key = sort_key,
        raise_errors = not asked.pass_errors,
        **kw,
        # Error handling parameters
        debug = asked.debug,
        section = "weaving"
    )

    # Output import file on stdout, in case the user would want to capture it.
    print(import_file)
    check_file(import_file)

    if asked.import_script_run:
        shell = os.environ["SHELL"]
        logger.info(f"Run import scripts with {shell}...")
        if asked.debug:
            subprocess.run([shell, import_file])

        else:
            try:
                subprocess.run([shell, import_file])
            except Exception as e:
                logger.error(e)
                sys.exit(ontoweaver.exceptions.SubprocessError.code)

    logger.info("Done")

if __name__ == "__main__":
    main()



import io
import yaml
import rdflib
import logging
import ontoweaver

def test_owladapter():
    ttl = """
@prefix : <http://fake.onto#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

:thing a owl:Class ;
        rdfs:label "thing" .

:source a owl:Class ;
        rdfs:subClassOf :thing ;
        rdfs:label "source" .

:target a owl:Class ;
        rdfs:subClassOf :thing ;
        rdfs:label "target" .

owl:ObjectProperty rdfs:subClassOf owl:thing .

:link a owl:ObjectProperty ;
        rdfs:subClassOf :thing ;
        rdfs:label "link" .

:prop a owl:DataProperty ;
        rdfs:subClassOf :thing ;
        rdfs:label "prop" .

:S0 a :source ;
        a owl:NamedIndividual ;
        rdfs:label "S0" ;
        :link :T0 .

:T0 a :target ;
        a owl:NamedIndividual ;
        rdfs:label "T0" ;
        :prop "data property" .
    """

    rdf_io = io.StringIO(ttl)
    g = rdflib.Graph()
    g.parse(rdf_io, format = "turtle")

    mapping = """
subject:
    map:
        id_from_element: label
        match_type_from_element: type
        match:
            - source:
                to_subject: source
            - target:
                to_subject: target
transformers:
    - map:
        element: link
        to_object: target
        via_relation: link
    - map:
        element: prop
        to_property: prop
metadata:
    - source: "Test"
    """
    config = yaml.full_load(mapping)

    lrg = ontoweaver.loader.LoadOWLGraph()
    local_nodes,local_edges = ontoweaver.load_extract(g, config, lrg, raise_errors = False)

    logging.debug("Nodes:")
    for n in local_nodes:
        logging.debug(repr(n))

    logging.debug("Edges:")
    for e in local_edges:
        logging.debug(repr(e))

    bc_nodes = [n.as_tuple() for n in local_nodes]
    bc_edges = [e.as_tuple() for e in local_edges]

    fnodes, fedges = ontoweaver.fusion.reconciliate(bc_nodes, bc_edges, reconciliate_sep=";")

    logging.debug("Nodes:")
    for n in fnodes:
        logging.debug(repr(n))

    logging.debug("Edges:")
    for e in fedges:
        logging.debug(repr(e))

    assert len(fnodes) == 2
    assert len(fedges) == 1

if __name__ == "__main__":
    test_owladapter()



"""
The exceptions hierarchy of what OntoWeaver may raise.

Classes in this file all have a static ``code`` variable,
which is used by the ``ontoweave`` command as an error return code.

The exceptions taxonomy is::
    
    OntoWeaverError
    ├─ AutoSchemaError
    ├─ ConfigError
    ├─ FeatureError
    ├─ ParsingError
    │  ├─ ParsingDeclarationsError
    │  │  ├─ CardinalityError
    │  │  └─ MissingDataError
    │  ├─ # ParsingPropertiesError
    │  │  └─ # MissingFieldError
    └─ RunError
       ├─ DeclarationError
       ├─ FileError
       │  ├─ FileAccessError
       │  └─ FileOverwriteError
       ├─ InputDataError
       ├─ SubprocessError
       └─ TransformerError
          ├─ DataValidationError
          ├─ InterfaceInheritanceError
          ├─ TransformerConfigError
          ├─ TransformerDataError
          ├─ TransformerInputError
          └─ TransformerInterfaceError
"""

class OntoWeaverError(Exception):
    """The base of all errors related to OntoWeaver."""
    code = 255

class AutoSchemaError(OntoWeaverError):
    """Your mapping is inconsistent with the ontology's taxonomy."""
    code = 79

class ConfigError(OntoWeaverError):
    """Your configuration has an error."""
    code = 78  # "bad config"

class FeatureError(OntoWeaverError):
    """This feature does not seem to exist."""
    code = 64  # alt "usage"

class ParsingError(OntoWeaverError):
    """I failed to parse a YAML file."""
    code = 65  # "data format"

class ParsingDeclarationsError(ParsingError):
    """Your mapping YAML file has an error."""
    code = 85

class CardinalityError(ParsingDeclarationsError):
    """Your mapping YAML file shows duplicated parameters."""
    code = 86

class MissingDataError(ParsingDeclarationsError):
    """Your mapping YAML file has missing parameter(s)."""
    code = 66  # "no input"

# class ParsingPropertiesError(ParsingError):

# class MissingFieldError(ParsingPropertiesError):

class RunError(OntoWeaverError):
    """Error while running OntoWeaver."""
    code = 70  # "internal"

class DeclarationError(RunError):
    """I cannot run your mapping, it has logical inconsistency."""
    code = 87

class FileError(RunError):
    """I have a problem with a file."""
    code = 72  # "OS file"

class FileAccessError(FileError):
    """I cannot read or write in a file."""
    code = 77  # "no perm"

class FileOverwriteError(FileError):
    """I cannot overwrite a file."""
    code = 126  # alt "no perm"

class InputDataError(RunError):
    """Some input data is inconsistent."""
    code = 132  # "illegal"

class TransformerError(RunError):
    """A transformer cannot run properly."""
    code = 200

class SubprocessError(RunError):
    """A subprocess ended on an error."""
    code = 128  # "bad exit"

class DataValidationError(TransformerError):
    """Some data validation do not pass."""
    code = 76  # "protocol"

class InterfaceInheritanceError(TransformerError):
    """The declaration of your user-made transformer is wrong."""
    code = 201

class TransformerConfigError(TransformerError):
    """Some parameters of your transformer are inconsistent."""
    code = 202

class TransformerDataError(TransformerError):
    """A transformer cannot process its data."""
    code = 203

class TransformerInputError(TransformerError):
    """The data a transformer transforms is inconsistent."""
    code = 204

class TransformerInterfaceError(TransformerError):
    """You used a transformer the wrong way."""
    code = 205




""" An empty module for hosting the element classes created on-the-fly.
"""
import sys
# from . import base

# Allow accessing all ontoweaver.Item classes defined in this module.
# all = base.All(sys.modules[__name__]) # FIXME circular import from base




def test_full_example():
    import logging
    import ontoweaver
    import biocypher
    import re
    import os

    logging.basicConfig(level=logging.DEBUG)

    dir = "full_example"

    data_mapping = {f"tests/{dir}/data.csv": f"tests/{dir}/mapping.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="none")

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges), reconciliate_sep=",")

    bc = biocypher.BioCypher(
        biocypher_config_path = f"tests/{dir}/config.yaml",
        schema_config_path = f"tests/{dir}/schema.yaml"
    )
    bc.show_ontology_structure()

    bc.write_nodes( fnodes )
    bc.write_edges( fedges )
    path = bc.write_import_call()

    outdir = os.path.dirname(path)
    with open(f"{outdir}/biocypher.ttl", 'r') as fd:
        ttl = fd.read()

    assert("S1" in ttl)
    assert("S2" in ttl)
    assert("GA" in ttl)
    assert("GO" in ttl)
    assert(":myedge" in ttl)

if __name__ == "__main__":
    test_full_example()




import rdflib
import ontoweaver
import pandas as pd

from . import testing_functions

def test_loader():
    lpf = ontoweaver.loader.LoadPandasFile()
    lpf.load(["./tests/simplest/data.csv"])
    lpf.load(["./tests/custom_transformer/data.tsv"])

    data = [1,2,3]
    df = pd.DataFrame(data)
    lpd = ontoweaver.loader.LoadPandasDataframe()
    lpd.load([df])

    lrf = ontoweaver.loader.LoadOWLFile()
    lrf.load(["./tests/test_preprocessing_ontology/OIM_test_preprocessing.owl"])

    g = rdflib.Graph()
    g.parse("./tests/test_preprocessing_ontology/bc_OIM_test_preprocessing.owl")
    lrg = ontoweaver.loader.LoadOWLGraph()
    lrg.load([g])


def test_multi_files():
    testing_functions.create_parquet_files()
    lpf = ontoweaver.loader.LoadPandasFile()
    lpf.load(["./tests/parquets/numbers_part-1.parquet", "./tests/parquets/numbers_part-2.parquet"])


def test_multi_files_glob():
    testing_functions.create_parquet_files()
    lpf = ontoweaver.loader.LoadPandasFile()
    lpf.load(["./tests/parquets/numbers_*.parquet"])




import io
import yaml
import logging
import ontoweaver

def test_from_file():
    mapping = """
row:
   map:
       column: /table/tbody/tr/td[1]
       to_subject: variant
transformers:
    - map:
        column: /table/tbody/tr/td[2]
        to_object: patient
        via_relation: patient_has_variant
    - map:
        column: /table/tbody/tr/td[3]
        to_property: age
        for_object: patient
metadata:
    - origin: somewhere
    """
    config = yaml.full_load(mapping)

    parser = ontoweaver.mapping.YamlParser(config)
    mapper = parser()

    lxs = ontoweaver.loader.LoadXMLFile()
    local_nodes,local_edges = ontoweaver.load_extract("tests/xml/friends.xml", config, lxs, raise_errors = False)

    for n in local_nodes:
        logging.debug(repr(n))
    for e in local_edges:
        logging.debug(repr(e))

    assert len(local_nodes) == 6
    assert len(local_edges) == 3


def test_xpath():
    xml="""
<table>
    <caption>Friends</caption>
    <thead>
        <tr><th>variant_id</th><th>patient</th><th>age</th></tr>
    </thead>
    <tbody>
        <tr><td>1</td><td>B</td><td>12</td></tr>
        <tr><td>0</td><td>A</td><td>23</td></tr>
        <tr><td>2</td><td>C</td><td>34</td></tr>
    </tbody>
</table>
    """

    mapping = """
row:
   map:
       element: /table/tbody/tr/td[1]
       to_subject: variant
transformers:
    - map:
        element: /table/tbody/tr/td[2]
        to_object: patient
        via_relation: patient_has_variant
    - map:
        element: /table/tbody/tr/td[3]
        to_property: age
        for_object: patient
metadata:
    - origin: somewhere
    """
    config = yaml.full_load(mapping)

    parser = ontoweaver.mapping.YamlParser(config)
    mapper = parser()

    lxs = ontoweaver.loader.LoadXMLString()
    local_nodes,local_edges = ontoweaver.load_extract(xml, config, lxs, raise_errors = False)

    for n in local_nodes:
        logging.debug(repr(n))
    for e in local_edges:
        logging.debug(repr(e))

    assert len(local_nodes) == 6
    assert len(local_edges) == 3

if __name__ == "__main__":
    test_xpath()



import logging

import ontoweaver

def test_append():

    sep = ";"
    merge = ontoweaver.merge.dictry.Append(sep)

    k = ontoweaver.base.Node()

    merge(k, {"p1":"x"},{"p2":"y"} )
    assert( merge.get() == {"p1":"x", "p2":"y"} )

    merge.reset()
    merge(k, {"p1":"x"},{} )
    assert( merge.get() == {"p1":"x"} )

    merge.reset()
    merge(k, {"p1":"x", "p2":"y"},{} )
    assert( merge.get() == {"p1":"x", "p2":"y"} )

    merge.reset()
    merge(k, {"p1":"x"},{"p1":"y"} )
    assert( "y" in merge.get()["p1"].split(sep) )
    assert( "x" in merge.get()["p1"].split(sep) )

    merge.reset()
    merge(k, {"p1":"abcd"},{"p1":"efgh"} )
    m = merge.get()
    assert( "abcd" in m["p1"].split(sep) )
    assert( "efgh" in m["p1"].split(sep) )

    merge.reset()
    merge(k, {"p1":"[abcd]"},{"p1":"[efgh]"} )
    m = merge.get()
    assert( "[abcd]" in m["p1"].split(sep) )
    assert( "[efgh]" in m["p1"].split(sep) )


if __name__ == "__main__":
    test_append()



""" Adapters that process tabular data.
"""
import logging

import pandas as pd
from typing import Optional
from collections.abc import Iterable
from enum import Enum, EnumMeta

from . import base
from . import transformer
from . import exceptions
from . import iterative
from . import validate

logger = logging.getLogger("ontoweaver")


class PandasAdapter(iterative.IterativeAdapter):
    """Interface for extracting data from a Pandas DataFrame with a simple mapping configuration based on declared types.

    The general idea is that each row of the table is mapped to a source node,
    and some column values are mapped to an edge leading to another node.
    Some other columns may also be mapped to properties of either a node or an edge.

    The class expect a configuration formed by three objects:

    - the type of the source node mapped for each row.
    - a dictionary mapping each column name to the type of the edge (which contains the type of both the source and target node),
    - a dictionary mapping each (node or edge) type to another dictionary listing which column is extracted to which property.

    Note that, when using the `configure` mapping,
    types are created by default in the `ontoweaver.types` module,
    so that you may access the list of all declared types by using:

    - `ontoweaver.types.all.nodes()`,
    - `ontoweaver.types.all.node_fields()`,
    - `ontoweaver.types.all.edges()`,
    - `ontoweaver.types.all.edge_fields()`.
    """

    def __init__(self,
            df: pd.DataFrame,
            subject_transformer: base.Transformer,
            transformers: Iterable[base.Transformer],
            metadata: Optional[dict] = None,
            validator: Optional[validate.InputValidator] = None,
            type_affix: Optional[base.TypeAffixes] = base.TypeAffixes.suffix,
            type_affix_sep: Optional[str] = ":",
            parallel_mapping: int = 0,
            raise_errors = True,
        ):

        super().__init__(
            subject_transformer,
            transformers,
            metadata,
            validator,
            type_affix,
            type_affix_sep,
            parallel_mapping,
            raise_errors
        )

        # logger.info("DataFrame info:")
        # logger.info(df.info())
        logger.debug("Columns available in the loaded Pandas DataFrame:")
        for c in df.columns:
            logger.debug(f"\t`{c}`")
        # pd.set_option('display.max_rows', 30)
        # pd.set_option('display.max_columns', 30)
        # pd.set_option('display.width', 150)
        # logger.info("\n" + str(df))
        self.df = df


    def iterate(self):
        return self.df.iterrows()




def test_properties_metadata():
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "properties_metadata"

    expected_nodes = [
        ('0:variant', 'variant', {'version': '1.1', 'database_name': 'my_database'}),
        ('1:variant', 'variant', {'version': '2.2', 'database_name': 'my_database'}),
        ('2:variant', 'variant', {'version': '3.3', 'database_name': 'my_database'}),
        ('A:patient', 'patient', {'database_name': 'my_database', 'source_columns': 'patient'}),
        ('B:patient', 'patient', {'database_name': 'my_database', 'source_columns': 'patient'}),
        ('C:patient', 'patient', {'database_name': 'my_database', 'source_columns': 'patient'}),
        ('publicationA:publication', 'publication', {'journal': 'journalA', 'database_name': 'my_database', 'source_columns': 'publication'}),
        ('publicationB:publication', 'publication', {'journal': 'journalB', 'database_name': 'my_database', 'source_columns': 'publication'}),
        ('publicationC:publication', 'publication', {'journal': 'journalC', 'database_name': 'my_database', 'source_columns': 'publication'}),
    ]

    expected_edges = [
        ('', '0:variant', 'A:patient', 'patient_has_variant', {'database_name': 'my_database'}),
        ('', '0:variant', 'publicationA:publication', 'publication_to_variant', {'database_name': 'my_database'}),
        ('', '1:variant', 'B:patient', 'patient_has_variant', {'database_name': 'my_database'}),
        ('', '1:variant', 'publicationB:publication', 'publication_to_variant', {'database_name': 'my_database'}),
        ('', '2:variant', 'C:patient', 'patient_has_variant', {'database_name': 'my_database'}),
        ('', '2:variant', 'publicationC:publication', 'publication_to_variant', {'database_name': 'my_database'}),
    ]


    data_mapping = {f"tests/{directory_name}/data.csv" : f"tests/{directory_name}/mapping.yaml" }

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix")

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges), reconciliate_sep=",")

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_properties_metadata()



def test_output_validation():
    from . import testing_functions
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "output_validation"

    expected_nodes = [
        ('1:variant', 'variant', {}),
        ('2:variant', 'variant', {}),
        ('3:variant', 'variant', {}),
        ('A:patient', 'patient', {'version': ''}),
        ('B:patient', 'patient', {'version': 'Correct'}),
    ]

    expected_edges = [
        ('', '1:variant', 'B:patient', 'patient_has_variant', {}),
        ('', '2:variant', 'A:patient', 'patient_has_variant', {}),
    ]

    data_mapping = {f"tests/{directory_name}/data.csv" : f"tests/{directory_name}/mapping.yaml" }

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix", validate_output=True, raise_errors=False)

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges), reconciliate_sep=",")

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_output_validation()



import logging
import yaml
import io
import pandas as pd

import ontoweaver


def test_transformer_split_match_2():

    data="""source	target	source_genesymbol	target_genesymbol	is_directed	is_stimulation	is_inhibition	consensus_direction	consensus_stimulation	consensus_inhibition	sources	references	omnipath	kinaseextra	ligrecextra	pathwayextra	mirnatarget	dorothea	collectri	tf_target	lncrna_mrna	tf_mirna	small_molecule	dorothea_curated	dorothea_chipseq	dorothea_tfbs	dorothea_coexp	dorothea_level	type	curation_effort	extra_attrs	evidences	ncbi_tax_id_source	entity_type_source	ncbi_tax_id_target	entity_type_target
htr2a-as1	P12004	htr2a-as1	PCNA	True	False	False	False	False	False	ncRDeathDB	ncRDeathDB:17932748	False	False	False	False	False	False	False	False	True	False	False						lncrna_post_transcriptional	1	"blabla"	9606	lncrna	9606	protein
P48281	Q9JMA7	Vdr	Cyp3a41a; Cyp3a41b	True	True	False	True	True	False	ExTRI_CollecTRI;HTRI_CollecTRI;HTRIdb;HTRIdb_DoRothEA;NTNU.Curated_CollecTRI;SIGNOR_CollecTRI;TRRUST;TRRUST_CollecTRI;TRRUST_DoRothEA	CollecTRI:11723248;CollecTRI:11991950;CollecTRI:12147248;	False	False	False	False	False	True	True	True	False	False	False	True	False	False	False	A;D	transcriptional	21	"blabla"	"blabli"	10090	protein	10090	protein"""

    csv = io.StringIO(data)
    table = pd.read_csv(csv, sep = '\t')

    mapping = """
    row:
        rowIndex:
            to_subject: Variant
    transformers:
        - split:
            separator: "; "
            id_from_column: target_genesymbol
            match_type_from_column: entity_type_target
            match:
              - protein:
                    to_object: target_gene
                    final_type: gene
                    via_relation: transcript_to_gene_relationship
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="suffix")
    fnodes, fedges = ontoweaver.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    for n in fnodes:
        logging.debug(n)

    assert len(fnodes) == 4


def test_transformer_split_match():

    logging.debug("Load data...")

    # Do not add newlines or spaces here
    # or else the parsing will be wrong.
    data = """Patient,Type
P1;P2,T1
P3;P4,T2
P5;P6,T1
P7,T2
P0,T0"""
    csv = io.StringIO(data)
    table = pd.read_csv(csv)

    mapping = """
    row:
        rowIndex:
            to_subject: Variant
    transformers:
        - split:
            separator: ";"
            id_from_column: Patient
            match_type_from_column: Type
            match:
                - T1:
                    to_object: T1
                    via_relation: has
                - T2:
                    to_object: T2
                    via_relation: has
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="suffix")
    fnodes, fedges = ontoweaver.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    for n in fnodes:
        logging.debug(n)

    assert len(fnodes) == 12


def test_transformer_split_string_properties_2():

    data="""source	target	source_genesymbol	target_genesymbol	is_directed	is_stimulation	is_inhibition	consensus_direction	consensus_stimulation	consensus_inhibition	sources	references	omnipath	kinaseextra	ligrecextra	pathwayextra	mirnatarget	dorothea	collectri	tf_target	lncrna_mrna	tf_mirna	small_molecule	dorothea_curated	dorothea_chipseq	dorothea_tfbs	dorothea_coexp	dorothea_level	type	curation_effort	extra_attrs	evidences	ncbi_tax_id_source	entity_type_source	ncbi_tax_id_target	entity_type_target
htr2a-as1	P12004	htr2a-as1	PCNA	True	False	False	False	False	False	ncRDeathDB	ncRDeathDB:17932748	False	False	False	False	False	False	False	False	True	False	False						lncrna_post_transcriptional	1	"blabla"	9606	lncrna	9606	protein
P48281	Q9JMA7	Vdr	Cyp3a41a; Cyp3a41b	True	True	False	True	True	False	ExTRI_CollecTRI;HTRI_CollecTRI;HTRIdb;HTRIdb_DoRothEA;NTNU.Curated_CollecTRI;SIGNOR_CollecTRI;TRRUST;TRRUST_CollecTRI;TRRUST_DoRothEA	CollecTRI:11723248;CollecTRI:11991950;CollecTRI:12147248;	False	False	False	False	False	True	True	True	False	False	False	True	False	False	False	A;D	transcriptional	21	"blabla"	"blabli"	10090	protein	10090	protein"""

    csv = io.StringIO(data)
    table = pd.read_csv(csv, sep = '\t')

    logging.debug("Load mappings...")

    mapping = """
    row:
        map:
            column: source
            to_subject: source
    transformers:
        - map:
            column: target
            to_object: target
            via_relation: source_has_target
        - split:
            column: target_genesymbol
            separator: "; "
            to_property: genesymbol
            for_object: target
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="none")
    fnodes, fedges = ontoweaver.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    for n in fnodes:
        logging.debug(n)

    for n in fnodes:
        if n[0] == 'Q9JMA7':
            assert 'genesymbol' in n[2]
            assert n[2]['genesymbol'] != ''
            assert '|' in n[2]['genesymbol']


def test_transformer_split_string_properties():

    logging.debug("Load data...")

    # Do not add newlines or spaces here
    # or else the parsing will be wrong.
    data = """Patient,Variant,Source
P1,V1-1,"S0; S1"
P1,V1-2,S1
P2,V2-1,"S2; S4"
P2,V2-2,S3"""
    csv = io.StringIO(data)
    table = pd.read_csv(csv)

    logging.debug("Load mappings...")

    mapping = """
    row:
        map:
            column: Variant
            to_subject: variant
    transformers:
        - map:
            column: Patient
            to_object: patient
            via_relation: patient_has_variant
        - split:
            column: Source
            separator: "; "
            to_property: source
            for_object: variant
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="none")
    fnodes, fedges = ontoweaver.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    for n in fnodes:
        logging.debug(n)

    for n in fnodes:
        logging.debug(n)
        if n[1] == 'variant':
            assert "source" in n[2]


def test_transformer_split_string():

    logging.debug("Load data...")

    # Do not add newlines or spaces here
    # or else the parsing will be wrong.
    data = """Patient,Variant,Source
P1,V1-1,"S0,S1"
P1,V1-2,S1
P2,V2-1,"S2,S4"
P2,V2-2,S3"""
    csv = io.StringIO(data)
    table = pd.read_csv(csv)

    logging.debug("Load mappings...")

    mapping = """
    row:
        map:
            column: Variant
            to_subject: variant
    transformers:
        - map:
            column: Patient
            to_object: patient
            via_relation: patient_has_variant
        - split:
            column: Source
            separator: ","
            to_object: source
            via_relation: has_source
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="none")
    fnodes, fedges = ontoweaver.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    for n in fnodes:
        logging.debug(n)
    assert len(fnodes) == 11


def test_transformer_split_list():

    logging.debug("Load data...")

    table = pd.DataFrame({
        'Patient': ['P1', 'P1', 'P2', 'P2'],
        'Variant': ['V1-1', 'V1-2', 'V2-1', 'V2-2'],
        'Source' : [['S0', 'S1'], 'S1', ['S2', 'S4'], 'S3']
    })

    logging.debug("Load mappings...")

    mapping = """
    row:
        map:
            column: Variant
            to_subject: variant
    transformers:
        - map:
            column: Patient
            to_object: patient
            via_relation: patient_has_variant
        - split:
            column: Source
            separator: ","
            to_object: source
            via_relation: has_source
    """

    map = yaml.safe_load(mapping)

    logging.debug("Run the adapter...")
    nodes, edges = ontoweaver.extract_table(table, map, affix="none")
    fnodes, fedges = ontoweaver.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    for n in fnodes:
        logging.debug(n)
    assert len(fnodes) == 11


if __name__ == "__main__":
    test_transformer_split_string()



